# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import datetime
from unittest import TestCase

import ddt
import time
from hamcrest import assert_that, equal_to, instance_of, only_contains, \
    raises, contains_string, is_in, has_items, none, calling

from storops.exception import UnityResourceNotFoundError, \
    UnityHostNameInUseError, UnityActionNotAllowedError, \
    UnityPolicyInvalidParametersError, UnitySnapScheduleNameInUseError
from storops.lib.resource import ResourceList
from storops.unity.enums import EnclosureTypeEnum, DiskTypeEnum, HealthEnum, \
    HostTypeEnum, ServiceLevelEnum, ServiceLevelEnumList, \
    StorageResourceTypeEnum, DNSServerOriginEnum, TierTypeEnum, \
    RaidTypeEnum, RaidStripeWidthEnum, StoragePoolTypeEnum, DiskTypeEnumList, \
    SpeedValuesEnum, ConnectorTypeEnum, FeatureStateEnum, \
    InterfaceConfigModeEnum, IpProtocolVersionEnum, SNMPAuthProtocolEnum, \
    SNMPPrivacyProtocolEnum
from storops.unity.resource.alert import UnityAlertList
from storops.unity.resource.alert_config_snmp import \
    UnityAlertConfigSNMPTarget, UnityAlertConfigSNMPTargetList
from storops.unity.resource.cifs_server import UnityCifsServerList
from storops.unity.resource.cifs_share import UnityCifsShareList, \
    UnityCifsShare
from storops.unity.resource.disk import UnityDisk
from storops.unity.resource.dns_server import UnityFileDnsServerList
from storops.unity.resource.filesystem import UnityFileSystemList, \
    UnityFileSystem
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.host import UnityHostInitiator, \
    UnityHostInitiatorList, UnityHost, UnityHostList
from storops.unity.resource.import_session import UnityImportSessionList
from storops.unity.resource.interface import UnityFileInterfaceList
from storops.unity.resource.job import UnityJobList
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.lun import UnityLunList
from storops.unity.resource.metric import UnityMetricQueryResultList, \
    UnityMetricRealTimeQueryList
from storops.unity.resource.move_session import UnityMoveSessionList
from storops.unity.resource.nas_server import UnityNasServer, \
    UnityNasServerList
from storops.unity.resource.nfs_server import UnityNfsServerList
from storops.unity.resource.nfs_share import UnityNfsShareList
from storops.unity.resource.pool import UnityPoolList, \
    RaidGroupParameter, UnityPool
from storops.unity.resource.port import UnityFcPortList
from storops.unity.resource.port import UnityIpPortList, \
    UnityEthernetPortList, UnityIscsiPortalList
from storops.unity.resource.snap import UnitySnapList
from storops.unity.resource.snap_schedule import UnitySnapScheduleRule
from storops.unity.resource.sp import UnityStorageProcessor, \
    UnityStorageProcessorList
from storops.unity.resource.system import UnitySystemList, UnitySystem, \
    UnityDpeList, UnityDpe, UnityVirusChecker, UnityVirusCheckerList, \
    UnityBasicSystemInfo, UnityBasicSystemInfoList, UnitySystemTime, \
    UnityNtpServer, UnityDae, UnityFeature, UnityMgmtInterfaceList, \
    UnitySystemCapacityList, \
    UnitySystemTierCapacity, UnityIscsiNodeList
from storops.unity.resource.vmware import UnityCapabilityProfileList
from storops_test.unity.rest_mock import t_rest, patch_rest, t_unity

__author__ = 'Cedric Zhuang'


@ddt.ddt
class UnitySystemTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        system = UnitySystem(cli=t_rest())
        assert_that(system.existed, equal_to(True))
        assert_that(system.health, instance_of(UnityHealth))
        assert_that(system.name, equal_to('FNM00151200215'))
        assert_that(system.model, equal_to('Unity 500'))
        assert_that(system.serial_number, equal_to('FNM00151200215'))
        assert_that(system.internal_model, equal_to('OBERON 25 DRIVE CHASSIS'))
        assert_that(system.platform, equal_to('Oberon_DualSP'))
        assert_that(system.mac_address, equal_to('08:00:1B:FF:EA:CD'))
        assert_that(system.is_eula_accepted, equal_to(True))
        assert_that(system.is_upgrade_complete, equal_to(True))
        assert_that(system.is_auto_failback_enabled, equal_to(True))
        assert_that(system.current_power, equal_to(469))
        assert_that(system.avg_power, equal_to(474))

    @patch_rest
    def test_init_by_ip(self):
        system = UnitySystem('10.244.223.66', 'admin', 'Password123!')
        assert_that(system.model, equal_to('Unity 500'))

    @patch_rest
    def test_get_all(self):
        systems = UnitySystemList(cli=t_rest())
        assert_that(len(systems), equal_to(1))

    @patch_rest
    def test_get_sp_all(self):
        unity = t_unity()
        sps = unity.get_sp()
        assert_that(sps, instance_of(UnityStorageProcessorList))
        assert_that(len(sps), equal_to(2))

    @patch_rest
    def test_get_spa(self):
        unity = t_unity()
        sp = unity.get_sp('spa')
        assert_that(sp, instance_of(UnityStorageProcessor))
        assert_that(sp.get_id(), equal_to('spa'))

    @patch_rest
    def test_get_capability_profile_service_levels(self):
        unity = t_unity()
        level_s = ServiceLevelEnum.SILVER
        level_p = ServiceLevelEnum.PLATINUM
        cp = unity.get_capability_profile(service_levels=[level_s, level_p])
        assert_that(cp, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp), 2)

        # use Enum or EnumList are both ok
        cp2 = unity.get_capability_profile(service_levels=level_s)
        assert_that(cp2, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp2), 1)
        assert_that(cp2[0].existed, equal_to(True))

        level_enum_list = ServiceLevelEnumList.parse([level_s, level_p])
        assert_that(level_enum_list, instance_of(ServiceLevelEnumList))
        cp3 = unity.get_capability_profile(service_levels=level_enum_list)
        assert_that(cp3, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp3), 2)
        assert_that(cp3[0].existed, equal_to(True))

    @patch_rest
    def test_get_capability_profile_usage_tags(self):
        unity = t_unity()
        tag = "capacity"
        cp = unity.get_capability_profile(usage_tags=tag)
        assert_that(cp, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp), 1)
        assert_that(cp[0].existed, equal_to(True))
        assert_that(tag, is_in(cp[0].usage_tags))

        cp2 = unity.get_capability_profile(usage_tags=[tag])
        assert_that(cp2, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp2), 1)
        assert_that(cp2[0].existed, equal_to(True))
        assert_that(tag, is_in(cp2[0].usage_tags))

        # None tags will not pass to rest url
        # it's same with get_capability_profile()
        cp3 = unity.get_capability_profile(usage_tags=None)
        assert_that(cp3, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp3), 2)
        assert_that(cp3[0].existed, equal_to(True))

    @patch_rest
    def test_get_capability_profile_not_found(self):
        unity = t_unity()
        cp = unity.get_capability_profile(service_levels=[5])
        assert_that(cp, instance_of(UnityCapabilityProfileList))
        assert_that(len(cp), equal_to(0))

    @patch_rest
    def test_get_sp_by_name(self):
        unity = t_unity()
        sp = unity.get_sp(name='SP A')
        assert_that(sp, instance_of(UnityStorageProcessor))
        assert_that(sp.id, equal_to('spa'))

    @patch_rest
    def test_get_lun_list(self):
        unity = t_unity()
        lun_list = unity.get_lun()
        assert_that(lun_list, instance_of(UnityLunList))
        assert_that(len(lun_list), equal_to(5))

    @patch_rest
    def test_create_host(self):
        unity = t_unity()
        host = unity.create_host("Hello")
        assert_that(host, instance_of(UnityHost))
        assert_that(host.id, equal_to('Host_11'))

    @patch_rest
    def test_create_host_existed(self):
        unity = t_unity()

        def f():
            # the 'flocker-3' is the Host_10 name
            unity.create_host("flocker-3")

        assert_that(f, raises(UnityHostNameInUseError))

    @patch_rest
    def test_get_portal_list(self):
        unity = t_unity()
        portals = unity.get_iscsi_portal()
        assert_that(portals, instance_of(UnityIscsiPortalList))

    @patch_rest
    def test_create_iscsi_portal(self):
        unity = t_unity()
        portal = unity.create_iscsi_portal(
            ethernet_port='spa_eth3', ip="10.244.213.244",
            netmask="255.255.255.0", vlan=133, gateway="10.244.213.1")
        assert_that(portal.id, equal_to('if_4'))

    @patch_rest
    def test_get_ethernet_list(self):
        unity = t_unity()
        ports = unity.get_ethernet_port()
        assert_that(ports, instance_of(UnityEthernetPortList))
        assert_that(len(ports), equal_to(8))

    @patch_rest
    def test_get_host_list(self):
        unity = t_unity()
        hosts = unity.get_host()
        assert_that(hosts, instance_of(UnityHostList))
        assert_that(len(hosts), equal_to(7))

    @patch_rest
    def test_get_initiators(self):
        unity = t_unity()
        initiators = unity.get_initiator()
        assert_that(initiators, instance_of(UnityHostInitiatorList))
        assert_that(len(initiators), equal_to(4))

    @patch_rest
    def test_get_initiator_by_id(self):
        unity = t_unity()
        initiator = unity.get_initiator(_id="HostInitiator_2")
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.id, equal_to("HostInitiator_2"))

    @patch_rest
    def test_get_initiator_by_name(self):
        unity = t_unity()
        wwn = "50:00:14:40:47:B0:0C:44:50:00:14:42:D0:0C:44:10"
        filtered = unity.get_initiator(initiator_id=wwn)
        assert_that(len(filtered), equal_to(1))
        assert_that(filtered, instance_of(UnityHostInitiatorList))
        initiator = filtered.first_item
        assert_that(initiator, instance_of(UnityHostInitiator))
        assert_that(initiator.existed, equal_to(True))
        assert_that(initiator.initiator_id, equal_to(wwn))

    @patch_rest
    def test_get_pools(self):
        unity = t_unity()
        pools = unity.get_pool()
        assert_that(pools, instance_of(UnityPoolList))
        assert_that(len(pools), equal_to(2))

    @patch_rest
    def test_create_pool(self):
        unity = t_unity()
        disk_group = unity.get_disk_group(_id='dg_15')
        raid_group_0 = RaidGroupParameter(
            disk_group=disk_group,
            disk_num=3, raid_type=RaidTypeEnum.RAID5,
            stripe_width=RaidStripeWidthEnum.BEST_FIT)
        raid_groups = [raid_group_0]
        pool = unity.create_pool(
            name='test_pool', description='Unity test pool.',
            raid_groups=raid_groups, alert_threshold=15,
            is_harvest_enabled=True, is_snap_harvest_enabled=True,
            pool_harvest_high_threshold=80, pool_harvest_low_threshold=40,
            snap_harvest_high_threshold=80, snap_harvest_low_threshold=40,
            is_fast_cache_enabled=True, is_fastvp_enabled=True,
            pool_type=StoragePoolTypeEnum.DYNAMIC)
        assert_that(pool, instance_of(UnityPool))

    @patch_rest
    def test_get_snaps_all(self):
        unity = t_unity()
        snaps = unity.get_snap()
        assert_that(snaps, instance_of(UnitySnapList))
        assert_that(len(snaps), equal_to(3))

    @patch_rest
    def test_get_snap_by_name(self):
        unity = t_unity()
        snap = unity.get_snap(name='2016-03-15_10:56:08')
        assert_that(snap.name, equal_to('2016-03-15_10:56:08'))
        assert_that(snap.existed, equal_to(True))

    @patch_rest
    def test_get_nas_servers(self):
        unity = t_unity()
        nas_servers = unity.get_nas_server()
        assert_that(nas_servers, instance_of(UnityNasServerList))
        assert_that(len(nas_servers), equal_to(3))

    @patch_rest
    def test_create_nas_server(self):
        unity = t_unity()
        sp = unity.get_sp(_id='spa')
        pool = unity.get_pool(_id='pool_1')
        nas_server = unity.create_nas_server('nas3', sp, pool)
        assert_that(nas_server.existed, equal_to(True))

    @patch_rest
    def test_auto_balance_sp_one_sp(self):
        unity = t_unity()

        @patch_rest(output='auto_balance_sp_one_sp.json')
        def inner():
            sp = unity._auto_balance_sp()
            assert_that(sp.get_id(), equal_to('spa'))

        unity._auto_balance_sp()
        inner()

    @patch_rest
    def test_auto_balance_sp_to_spb(self):
        unity = t_unity()
        sp = unity._auto_balance_sp()
        assert_that(sp.get_id(), equal_to('spb'))

    @patch_rest
    def test_auto_balance_sp_to_spa(self):
        unity = t_unity()

        @patch_rest(output='auto_balance_sp_to_spa.json')
        def inner():
            sp = unity._auto_balance_sp()
            assert_that(sp.get_id(), equal_to('spa'))

        unity._auto_balance_sp()
        inner()

    @patch_rest
    def test_get_ip_ports(self):
        unity = t_unity()
        ip_ports = unity.get_ip_port()
        assert_that(ip_ports, instance_of(UnityIpPortList))
        assert_that(len(ip_ports), equal_to(8))

    @patch_rest
    def test_get_file_interface(self):
        unity = t_unity()
        fi_list = unity.get_file_interface()
        assert_that(fi_list, instance_of(UnityFileInterfaceList))
        assert_that(len(fi_list), equal_to(1))

    @patch_rest
    def test_get_cifs_server(self):
        unity = t_unity()
        cifs_servers = unity.get_cifs_server()
        assert_that(cifs_servers, instance_of(UnityCifsServerList))
        assert_that(len(cifs_servers), equal_to(1))

    @patch_rest
    def test_get_nfs_server(self):
        unity = t_unity()
        nfs_servers = unity.get_nfs_server()
        assert_that(nfs_servers, instance_of(UnityNfsServerList))
        assert_that(len(nfs_servers), equal_to(1))

    @patch_rest
    def test_get_dns_server(self):
        unity = t_unity()
        dns_servers = unity.get_dns_server()
        assert_that(dns_servers, instance_of(UnityFileDnsServerList))
        assert_that(len(dns_servers), equal_to(1))

    @patch_rest
    def test_get_file_system(self):
        unity = t_unity()
        filesystems = unity.get_filesystem()
        assert_that(filesystems, instance_of(UnityFileSystemList))
        assert_that(len(filesystems), equal_to(3))

    @patch_rest
    def test_get_cifs_share(self):
        unity = t_unity()
        shares = unity.get_cifs_share()
        assert_that(shares, instance_of(UnityCifsShareList))
        assert_that(len(shares), equal_to(1))

    @patch_rest
    def test_get_nfs_share(self):
        unity = t_unity()
        shares = unity.get_nfs_share()
        assert_that(shares, instance_of(UnityNfsShareList))
        assert_that(len(shares), equal_to(2))

    @patch_rest
    def test_get_host_by_address_found(self):
        unity = t_unity()
        host = unity.get_host(address='8.8.8.8')
        assert_that(host.type, equal_to(HostTypeEnum.SUBNET))

    @patch_rest
    def test_get_host_by_address_not_found(self):
        unity = t_unity()
        hosts = unity.get_host(address='8.8.8.9')
        assert_that(len(hosts), equal_to(0))

    @patch_rest
    def test_system_info(self):
        unity = t_unity()
        assert_that(unity.info, instance_of(UnityBasicSystemInfo))

    @patch_rest
    def test_system_get_cifs_share_by_name(self):
        unity = t_unity()
        cs = unity.get_cifs_share(name='cs1')
        assert_that(cs, instance_of(UnityCifsShare))
        assert_that(cs.name, equal_to('cs1'))

    @patch_rest
    def test_system_get_fs_by_name_not_found(self):
        def f():
            unity = t_unity()
            unity.get_filesystem(name='not_found')

        assert_that(f, raises(UnityResourceNotFoundError,
                              'UnityFileSystem:not_found'))

    @patch_rest
    def test_get_doc_enum_member(self):
        unity = t_unity()
        doc = unity.get_doc(HealthEnum.NON_RECOVERABLE)
        assert_that(doc, contains_string('OK But Minor Warning'))

    @patch_rest
    def test_get_doc_enum(self):
        unity = t_unity()
        doc = unity.get_doc(HealthEnum)
        assert_that(doc, contains_string('OK But Minor Warning'))

    @patch_rest
    def test_get_doc_resource(self):
        unity = t_unity()
        doc = unity.get_doc(unity.get_snap())
        assert_that(doc, contains_string(
            'For a file system or VMware NFS datastore'))

    @patch_rest
    def test_get_fc_port(self):
        unity = t_unity()
        fi_list = unity.get_fc_port()
        assert_that(fi_list, instance_of(UnityFcPortList))
        assert_that(len(fi_list), equal_to(12))

    @patch_rest(output='fc_port_not_supported.json')
    def test_get_fc_port_not_supported(self):
        unity = t_unity(version='4.1.2')
        fc = unity.get_fc_port()
        assert_that(len(fc), equal_to(0))

    @patch_rest
    def test_get_io_limit_policy_by_name(self):
        unity = t_unity()
        name = 'Density_1100_KBPS'
        policy = unity.get_io_limit_policy(name=name)
        assert_that(policy.name, equal_to(name))

    @patch_rest
    def test_get_io_limit_policy_all(self):
        unity = t_unity()
        policies = unity.get_io_limit_policy()
        assert_that(len(policies), equal_to(6))

    @patch_rest
    def test_create_kbps_policy(self):
        unity = t_unity()
        policy = unity.create_io_limit_policy(
            'max_kbps_1234', max_kbps=1234, description='storops')
        assert_that(policy.name, equal_to('max_kbps_1234'))
        setting = policy.io_limit_rule_settings[0]
        assert_that(setting.max_kbps, equal_to(1234))

    @patch_rest
    def test_create_burst_policy(self):
        unity = t_unity()
        policy = unity.create_io_limit_policy(
            'burst_policy', description='storops', max_iops=1000,
            max_kbps=20480, burst_rate=50, burst_time=10, burst_frequency=2)
        assert_that(policy.name, equal_to('burst_policy'))
        setting = policy.io_limit_rule_settings[0]
        assert_that(setting.burst_rate, equal_to(50))
        assert_that(setting.burst_time, equal_to(10))
        assert_that(setting.burst_frequency, equal_to(2))

    @patch_rest
    def test_create_cg(self):
        lun1 = UnityLun(cli=t_rest(), _id='sv_3339')
        lun2 = UnityLun(cli=t_rest(), _id='sv_3340')
        unity = t_unity()
        cg = unity.create_cg('Muse', lun_add=[lun1, lun2])
        assert_that(cg.name, equal_to('Muse'))
        assert_that(len(cg.luns), equal_to(2))

    @patch_rest
    def test_get_cg_list(self):
        cg_list = t_unity().get_cg()
        assert_that(len(cg_list), equal_to(2))

    @patch_rest
    def test_get_cg_by_name(self):
        cg = t_unity().get_cg(name='Nike')
        assert_that(cg.name, equal_to('Nike'))

    @patch_rest
    def test_get_cg_by_id(self):
        cg = t_unity().get_cg(_id='res_13')
        assert_that(cg.id, equal_to('res_13'))
        cg_type = StorageResourceTypeEnum.CONSISTENCY_GROUP
        assert_that(cg.type, equal_to(cg_type))

    @patch_rest
    def test_system_time(self):
        assert_that(t_unity().system_time.year, equal_to(2016))

    @patch_rest
    def test_set_system_time(self):
        unity = t_unity()
        st = unity.system_time
        ret = unity.set_system_time(new_time=st)
        assert_that(ret, instance_of(datetime.datetime))
        assert_that(ret.microsecond, equal_to(534000))

    @patch_rest
    def test_set_ntp_server(self):
        ret = t_unity().add_ntp_server('10.245.54.154', '10.245.54.153')
        assert_that(ret, has_items('10.245.54.152', '10.245.54.153'))

    @patch_rest
    def test_remove_ntp_server(self):
        ret = t_unity().remove_ntp_server('10.245.54.153')
        assert_that(ret, has_items('10.245.54.152', '10.245.54.153'))

    @patch_rest
    def test_clear_ntp_server(self):
        ret = t_unity().clear_ntp_server()
        assert_that(ret, has_items('10.245.54.152', '10.245.54.153'))

    @patch_rest
    def test_list_ntp_servers(self):
        assert_that(t_unity().ntp_server,
                    has_items('10.245.54.152', '10.245.54.153'))

    @patch_rest
    def test_delete_singleton(self):
        def f():
            t_unity()._ntp_server.delete()

        assert_that(f, raises(UnityActionNotAllowedError, 'not allowed'))

    @patch_rest
    def test_list_dns_server(self):
        dns_server = t_unity().dns_server
        assert_that(dns_server.id, equal_to('0'))
        assert_that(dns_server.origin, equal_to(DNSServerOriginEnum.DHCP))
        assert_that(dns_server.addresses,
                    has_items('10.245.177.15', '10.245.177.16'))

    @patch_rest
    def test_add_dns_server(self):
        ret = t_unity().add_dns_server('8.8.8.8', '9.9.9.9')
        assert_that(ret, has_items('10.245.177.15', '10.245.177.16'))

    @patch_rest
    def test_remove_dns_server(self):
        ret = t_unity().remove_dns_server('8.8.8.8', '9.9.9.9')
        assert_that(ret, has_items('10.245.177.15', '10.245.177.16'))

    @patch_rest
    def test_clear_dns_server(self):
        ret = t_unity().clear_dns_server()
        assert_that(ret, has_items('10.245.177.15', '10.245.177.16'))

    @patch_rest
    def test_get_link_aggregation_list(self):
        la_list = t_unity().get_link_aggregation()
        assert_that(len(la_list), equal_to(2))

    @patch_rest
    def test_get_disk(self):
        disks = t_unity().get_disk(tier_type=TierTypeEnum.EXTREME_PERFORMANCE)
        assert_that(len(disks), equal_to(4))

    @patch_rest
    def test_get_link_aggregation_by_name(self):
        cg = t_unity().get_link_aggregation(name='SP A Link Aggregation 2')
        assert_that(cg.name, equal_to('SP A Link Aggregation 2'))

    @patch_rest
    def test_get_link_aggregation_by_id(self):
        cg = t_unity().get_link_aggregation(_id='spa_la_2')
        assert_that(cg.id, equal_to('spa_la_2'))

    @patch_rest(output='link_aggregation_not_supported.json')
    def test_get_link_aggregation_not_supported(self):
        unity = t_unity(version='4.1.2')
        la = unity.get_link_aggregation()
        assert_that(len(la), equal_to(0))

    @patch_rest
    def test_get_file_port(self):
        unity = UnitySystem(cli=t_rest("4.1.0"))
        ports = unity.get_file_port()
        assert_that(len(ports), equal_to(6))

    @patch_rest
    def test_get_file_port_la_unsupported(self):
        unity = UnitySystem(cli=t_rest("4.0.0"))
        ports = unity.get_file_port()
        assert_that(len(ports), equal_to(8))

    @patch_rest
    def test_enable_performance_statistics(self):
        unity = UnitySystem('10.244.223.61')
        assert_that(unity.is_perf_stats_enabled(), equal_to(False))

        queries = unity.enable_perf_stats(1)
        assert_that(queries, instance_of(UnityMetricRealTimeQueryList))

        time.sleep(5)
        assert_that(unity.is_perf_stats_enabled(), equal_to(True))
        assert_that(unity._cli.curr_counter,
                    instance_of(UnityMetricQueryResultList))

        unity.disable_perf_stats()
        assert_that(unity.is_perf_stats_enabled(), equal_to(False))
        assert_that(unity._cli.curr_counter, none())
        assert_that(unity._cli.prev_counter, none())

    @patch_rest
    def test_enable_persist_perf_stats(self):
        unity = UnitySystem('10.244.223.61')
        assert_that(unity.is_perf_stats_persisted(), equal_to(False))

        unity.enable_persist_perf_stats()
        assert_that(unity.is_perf_stats_persisted(), equal_to(True))

        unity.disable_persist_perf_stats()
        assert_that(unity.is_perf_stats_persisted(), equal_to(False))

    @patch_rest
    def test_default_rsc_clz_list_with_perf_stats(self):
        rsc_list_collection = t_unity()._default_rsc_list_with_perf_stats()
        clz_list = ResourceList.get_rsc_clz_list(rsc_list_collection)
        assert_that(clz_list, has_items(UnityDisk, UnityLun, UnityFileSystem))

    @patch_rest
    def test_get_tenant(self):
        unity = t_unity()
        tenant = unity.get_tenant()
        assert_that(len(tenant), equal_to(3))

    @patch_rest
    def test_get_tenant_use_vlan(self):
        unity = t_unity()
        tenant = unity.get_tenant_use_vlan(4)
        assert_that(tenant.id, equal_to('tenant_4'))

    @patch_rest
    def test_get_tenant_use_vlan_not_found(self):
        unity = t_unity()
        tenant = unity.get_tenant_use_vlan(5)
        assert_that(tenant, equal_to(None))

    @patch_rest
    def test_create_tenant(self):
        unity = t_unity()
        unity.create_tenant(
            'test', uuid='173ca6c3-5952-427d-82a6-df88f49e3926',
            vlans=[3])

    @patch_rest(output='system/license_upload.json')
    def test_upload_license(self):
        unity = t_unity()
        unity.upload_license(license="license.lic")

    @patch_rest
    def test_get_iscsi_node(self):
        unity = t_unity()
        nodes = unity.get_iscsi_node()
        assert_that(nodes, instance_of(UnityIscsiNodeList))
        assert_that(len(nodes), equal_to(4))

    @patch_rest
    def test_get_mgmt_interface(self):
        unity = t_unity()
        interfaces = unity.get_mgmt_interface()
        assert_that(interfaces, instance_of(UnityMgmtInterfaceList))
        assert_that(len(interfaces), equal_to(1))

    @patch_rest
    def test_get_system_capacity(self):
        unity = t_unity()
        capacities = unity.get_system_capacity()
        assert_that(capacities, instance_of(UnitySystemCapacityList))
        assert_that(len(capacities), equal_to(1))

    @patch_rest
    def test_get_move_session(self):
        unity = t_unity()
        move_sessions = unity.get_move_session()
        assert_that(move_sessions, instance_of(UnityMoveSessionList))
        assert_that(len(move_sessions), equal_to(1))

    @patch_rest
    def test_get_import_session(self):
        unity = t_unity()
        sessions = unity.get_import_session()
        assert_that(sessions, instance_of(UnityImportSessionList))
        assert_that(len(sessions), equal_to(3))

    @patch_rest
    def test_create_alert_snmp_v3_config(self):
        unity = t_unity()
        alert_snmp_config = unity.create_alert_snmp_config(
            '10.10.10.111', username='test_username',
            auth_protocol=SNMPAuthProtocolEnum.MD5,
            priv_protocol=SNMPPrivacyProtocolEnum.AES,
            auth_password='auth_password_test')
        assert_that(alert_snmp_config, instance_of(
            UnityAlertConfigSNMPTarget))
        assert_that(alert_snmp_config.username, equal_to('test_username'))
        assert_that(alert_snmp_config.get_id(), equal_to('snmp_target_8'))
        assert_that(alert_snmp_config.address, equal_to('10.10.10.111'))

    @patch_rest
    def test_create_alert_snmp_v2_config(self):
        unity = t_unity()
        alert_snmp_config = unity.create_alert_snmp_config(
            '10.10.10.222', community='test_community')
        assert_that(alert_snmp_config, instance_of(
            UnityAlertConfigSNMPTarget))
        assert_that(alert_snmp_config.get_id(), equal_to(
            'snmp_target_9'))
        assert_that(alert_snmp_config.address, equal_to('10.10.10.222'))

    @patch_rest
    def test_create_invalid_snmp_config(self):
        def f():
            unity = t_unity()
            unity.create_alert_snmp_config('10.244.166.123')

        assert_that(f, raises(
            UnityPolicyInvalidParametersError,
            'One or more specified parameters are invalid.'))

    @patch_rest
    def test_query_alert_snmp_config(self):
        unity = t_unity()
        alert_snmp_config = unity.get_alert_snmp_config(
            'snmp_target_8')
        assert_that(alert_snmp_config, instance_of(
            UnityAlertConfigSNMPTarget))
        assert_that(alert_snmp_config.get_id(), equal_to(
            'snmp_target_8'))

    @patch_rest
    def test_query_by_ip(self):
        unity = t_unity()
        alert_snmp_configs = unity.get_alert_snmp_config(
            address='16.16.16.16')
        assert_that(alert_snmp_configs, instance_of(
            UnityAlertConfigSNMPTargetList))
        assert_that(len(alert_snmp_configs), equal_to(2))

    @patch_rest
    def test_query_all_alert_snmp_configs(self):
        unity = t_unity()
        alert_snmp_configs = unity.get_alert_snmp_config()
        assert_that(alert_snmp_configs, instance_of(
            UnityAlertConfigSNMPTargetList))
        assert_that(len(alert_snmp_configs), equal_to(4))

    @patch_rest
    def test_get_job(self):
        unity = t_unity()
        jobs = unity.get_job()
        assert_that(jobs, instance_of(UnityJobList))
        assert_that(len(jobs), equal_to(3))

    @patch_rest
    def test_get_alert(self):
        unity = t_unity()
        alerts = unity.get_alert()
        assert_that(alerts, instance_of(UnityAlertList))
        assert_that(len(alerts), equal_to(3))

    @ddt.data(
        {'rules': [UnitySnapScheduleRule.every_day([12])],
         'is_sync_replicated': None,
         'skip_sync_to_remote_system': None},
        {'rules': [UnitySnapScheduleRule.every_day([0])],
         'is_sync_replicated': True,
         'skip_sync_to_remote_system': True},
    )
    @patch_rest
    def test_create_snap_schedule(self, kwargs):
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        unity = t_unity()
        schedule = unity.create_snap_schedule(
            'snap-schedule-created-by-system',
            **kwargs
        )
        assert_that(schedule.get_id(), equal_to('snapSch_6'))

    @patch_rest
    def test_create_snap_schedule_existing(self):
        unity = t_unity()
        assert_that(
            calling(unity.create_snap_schedule).with_args(
                'snap-schedule-created-by-system-existing',
                [UnitySnapScheduleRule.every_month([15])]
            ),
            raises(UnitySnapScheduleNameInUseError)
        )


class UnityDpeTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        dpe = UnityDpe('dpe', cli=t_rest())
        assert_that(dpe.existed, equal_to(True))
        assert_that(dpe.health, instance_of(UnityHealth))
        assert_that(dpe.needs_replacement, equal_to(False))
        assert_that(dpe.slot_number, equal_to(0))
        assert_that(dpe.name, equal_to('DPE'))
        assert_that(dpe.manufacturer, equal_to(''))
        assert_that(dpe.model, equal_to('OBERON 25 DRIVE CHASSIS'))
        assert_that(dpe.emc_part_number, equal_to('100-542-901-05'))
        assert_that(dpe.emc_serial_number, equal_to('CF2CV150500005'))
        assert_that(dpe.vendor_part_number, equal_to(''))
        assert_that(dpe.vendor_serial_number, equal_to(''))
        assert_that(dpe.bus_id, equal_to(0))
        assert_that(dpe.current_power, equal_to(429))
        assert_that(dpe.avg_power, equal_to(397))
        assert_that(dpe.max_power, equal_to(429))
        assert_that(dpe.current_temperature, equal_to(26))
        assert_that(dpe.avg_temperature, equal_to(26))
        assert_that(dpe.max_temperature, equal_to(26))
        assert_that(dpe.current_speed, equal_to(12000000000))
        assert_that(dpe.max_speed, equal_to(12000000000))
        assert_that(dpe.parent_system, instance_of(UnitySystem))
        assert_that(dpe.enclosure_type,
                    equal_to(EnclosureTypeEnum.MIRANDA_12G_SAS_DPE))
        assert_that(dpe.drive_types,
                    only_contains(DiskTypeEnum.SAS, DiskTypeEnum.SAS_FLASH_2))

    @patch_rest
    def test_get_all(self):
        dpe_list = UnityDpeList(cli=t_rest())
        assert_that(len(dpe_list), equal_to(1))


class UnityVirusCheckerTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        checker = UnityVirusChecker('cava_2', cli=t_rest())
        assert_that(checker.existed, equal_to(True))
        assert_that(checker.nas_server, instance_of(UnityNasServer))
        assert_that(checker.is_enabled, equal_to(False))

    @patch_rest
    def test_get_all(self):
        checker_list = UnityVirusCheckerList(cli=t_rest())
        assert_that(len(checker_list), equal_to(1))


class UnityBasicSystemInfoTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        info = UnityBasicSystemInfo(t_rest())
        assert_that(info.id, equal_to('0'))
        assert_that(info.existed, equal_to(True))
        assert_that(info.name, equal_to('FNM00151200215'))
        assert_that(info.software_version, equal_to('4.1.0'))
        assert_that(info.earliest_api_version, equal_to('4.0'))
        assert_that(info.model, equal_to('Unity 500'))
        assert_that(info.api_version, equal_to("5.0"))

    @patch_rest
    def test_get_all(self):
        info_list = UnityBasicSystemInfoList(cli=t_rest())
        assert_that(len(info_list), equal_to(1))


class UnitySystemTimeTest(TestCase):
    @patch_rest
    def test_get_system_time(self):
        st = UnitySystemTime(t_rest()).time
        assert_that(st, instance_of(datetime.datetime))
        assert_that(st.year, equal_to(2016))
        assert_that(st.month, equal_to(11))
        assert_that(st.day, equal_to(14))
        assert_that(st.hour, equal_to(7))
        assert_that(st.minute, equal_to(23))
        assert_that(st.second, equal_to(53))


class UnityNtpServerTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        ntp_server = UnityNtpServer(cli=t_rest())
        assert_that(ntp_server.addresses,
                    has_items('10.245.54.152', '10.245.54.153'))


class UnityBatteryTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        batteries = t_unity().get_battery()
        assert_that(len(batteries), equal_to(2))
        battery0 = batteries[0]
        assert_that(battery0.id, equal_to("spa_bbu_0"))
        assert_that(battery0.health, instance_of(UnityHealth))
        assert_that(battery0.needs_replacement, equal_to(False))
        assert_that(battery0.parent, instance_of(dict))
        assert_that(battery0.slot_number, equal_to(0))
        assert_that(battery0.name, equal_to("SP A Battery 0"))
        assert_that(battery0.manufacturer, equal_to("ACBEL POLYTECH INC."))
        assert_that(battery0.model, equal_to("LITHIUM-ION, UNIVERSAL BOB"))
        assert_that(battery0.firmware_version, equal_to("073.91"))
        assert_that(battery0.emc_part_number, equal_to("078-000-128-02"))
        assert_that(battery0.emc_serial_number, equal_to("ACPJ5143800045"))
        assert_that(battery0.vendor_part_number, equal_to("SGD006-710G"))
        assert_that(battery0.vendor_serial_number, equal_to(""))
        assert_that(battery0.parent_storage_processor,
                    instance_of(UnityStorageProcessor))

    @patch_rest
    def test_get_nested_properties(self):
        batteries = t_unity().get_battery()
        assert_that(len(batteries), equal_to(2))
        battery0 = batteries[0]
        assert_that(battery0.parent_storage_processor.id, equal_to("spa"))
        assert_that(battery0.parent_storage_processor.name, equal_to("SP A"))


class UnityDaeTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        daes = t_unity().get_dae()
        assert_that(len(daes), equal_to(1))
        dae0 = daes[0]
        assert_that(dae0.id, equal_to("dae_0_1"))
        assert_that(dae0.enclosure_type,
                    equal_to(EnclosureTypeEnum.ANCHO_12G_SAS_DAE))
        assert_that(dae0.drive_types, instance_of(DiskTypeEnumList))
        assert_that(dae0.drive_types[0], DiskTypeEnum.NL_SAS)
        assert_that(dae0.health, instance_of(UnityHealth))
        assert_that(dae0.needs_replacement, equal_to(False))
        assert_that(dae0.parent, instance_of(dict))
        assert_that(dae0.slot_number, equal_to(1))
        assert_that(dae0.name, equal_to("DAE 0 1"))
        assert_that(dae0.manufacturer, equal_to(""))
        assert_that(dae0.model, equal_to("ANCHO LF 12G SAS DAE"))
        assert_that(dae0.emc_part_number, equal_to("100-900-000-04"))
        assert_that(dae0.emc_serial_number, equal_to("CF22W145100058"))
        assert_that(dae0.vendor_part_number, equal_to(""))
        assert_that(dae0.vendor_serial_number, equal_to(""))
        assert_that(dae0.bus_id, equal_to(0))
        assert_that(dae0.current_power, equal_to(110))
        assert_that(dae0.avg_power, equal_to(110))
        assert_that(dae0.max_power, equal_to(110))
        assert_that(dae0.current_temperature, equal_to(25))
        assert_that(dae0.avg_temperature, equal_to(25))
        assert_that(dae0.max_temperature, equal_to(25))
        assert_that(dae0.current_speed, equal_to(12000000000))
        assert_that(dae0.max_speed, equal_to(12000000000))
        assert_that(dae0.parent_system, instance_of(UnitySystem))


class UnityLccTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        lccs = t_unity().get_lcc()
        assert_that(len(lccs), equal_to(2))
        lcc = lccs[0]
        assert_that(lcc.id, equal_to("dae_0_1_lcc_a"))
        assert_that(lcc.health, instance_of(UnityHealth))
        assert_that(lcc.needs_replacement, equal_to(False))
        assert_that(lcc.parent, instance_of(dict))
        assert_that(lcc.slot_number, equal_to(0))
        assert_that(lcc.name, equal_to("DAE 0 1 Link Control Card A"))
        assert_that(lcc.manufacturer, equal_to(""))
        assert_that(lcc.model, equal_to("ANCHO 12G SAS LCC FRU ASSY"))
        assert_that(lcc.sas_expander_versions, equal_to(["2.20.0"]))
        assert_that(lcc.emc_part_number, equal_to("303-300-000C-02"))
        assert_that(lcc.emc_serial_number, equal_to("CF2W9145000032"))
        assert_that(lcc.vendor_part_number, equal_to(""))
        assert_that(lcc.vendor_serial_number, equal_to(""))
        assert_that(lcc.current_speed, equal_to(12000000000))
        assert_that(lcc.max_speed, equal_to(12000000000))
        assert_that(lcc.parent_dae, instance_of(UnityDae))


class UnityMemoryModuleTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        memory_modules = t_unity().get_memory_module()
        assert_that(len(memory_modules), equal_to(8))
        memory_module = memory_modules[0]
        assert_that(memory_module.id, equal_to("spa_mm_0"))
        assert_that(memory_module.health, instance_of(UnityHealth))
        assert_that(memory_module.needs_replacement, equal_to(False))
        assert_that(memory_module.slot_number, equal_to(0))
        assert_that(memory_module.name, equal_to("SP A Memory Module 0"))
        assert_that(memory_module.manufacturer, equal_to("Samsung"))
        assert_that(memory_module.model, equal_to("DDR4 SDRAM"))
        assert_that(memory_module.firmware_version, equal_to(""))
        assert_that(memory_module.size, equal_to(16))
        assert_that(memory_module.emc_part_number, equal_to("100-564-193-00"))
        assert_that(memory_module.emc_serial_number,
                    equal_to("80CE02151171BE4865"))
        assert_that(memory_module.vendor_part_number,
                    equal_to("M393A2G40DB0-CPB"))
        assert_that(memory_module.vendor_serial_number,
                    equal_to("71BE4865"))
        assert_that(memory_module.is_inserted, equal_to(True))
        assert_that(memory_module.parent_storage_processor,
                    instance_of(UnityStorageProcessor))


class UnityPowerSupplyTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        supplies = t_unity().get_power_supply()
        assert_that(len(supplies), equal_to(4))
        supply = supplies[0]

        assert_that(supply.id, equal_to("dae_0_1_ps_a0"))
        assert_that(supply.health, instance_of(UnityHealth))
        assert_that(supply.needs_replacement, equal_to(False))
        assert_that(supply.slot_number, equal_to(0))
        assert_that(supply.name, equal_to("DAE 0 1 Power Supply A0"))
        assert_that(supply.manufacturer, equal_to("ACBEL POLYTECH INC."))
        assert_that(supply.model, equal_to("Third Gen VE.400W, Dual +12V P/S"))
        assert_that(supply.firmware_version, equal_to("0421"))
        assert_that(supply.emc_serial_number, equal_to("AC7B7143200521"))
        assert_that(supply.vendor_part_number, equal_to("SGA001-710G"))
        assert_that(supply.vendor_serial_number, equal_to("AC7143202316"))
        assert_that(supply.emc_part_number, equal_to("071-000-553"))
        assert_that(supply.parent_dae, instance_of(UnityDae))
        assert_that(supply.storage_processor,
                    instance_of(UnityStorageProcessor))

    @patch_rest
    def test_get_nested_properties(self):
        supplies = t_unity().get_power_supply()
        assert_that(len(supplies), equal_to(4))

        supply0 = supplies[0]
        assert_that(supply0.parent_dae.id, equal_to('dae_0_1'))
        assert_that(supply0.parent_dae.name, equal_to('DAE 0 1'))
        assert_that(supply0.storage_processor.id, equal_to('spa'))
        assert_that(supply0.storage_processor.name, equal_to('SP A'))

        supply2 = supplies[2]
        assert_that(supply2.parent_dpe.id, equal_to('dpe'))
        assert_that(supply2.parent_dpe.name, equal_to('DPE'))


class UnitySasPortTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        ports = t_unity().get_sas_port()
        assert_that(len(ports), equal_to(4))
        port = ports[0]

        assert_that(port.id, equal_to("spa_sas0"))
        assert_that(port.health, instance_of(UnityHealth))
        assert_that(port.needs_replacement, equal_to(False))
        assert_that(port.name, equal_to("SP A SAS Port 0"))
        assert_that(port.port, equal_to(0))
        assert_that(port.current_speed, equal_to(SpeedValuesEnum._12Gbps))
        assert_that(port.connector_type,
                    equal_to(ConnectorTypeEnum.MINI_SAS_HD))
        assert_that(port.storage_processor, instance_of(UnityStorageProcessor))
        assert_that(port.parent_io_module, none())
        assert_that(port.parent_storage_processor,
                    instance_of(UnityStorageProcessor))


# TODO find a system with ssc
class UnitySscTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        sscs = t_unity().get_ssc()

        assert_that(len(sscs), equal_to(0))
        # ssc = sscs[0]
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))
        # assert_that(ssc.id, equal_to(""))


class UnitySsdTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        ssds = t_unity().get_ssd()

        ssd = ssds[0]
        assert_that(ssd.id, equal_to("spa_ssd"))
        assert_that(ssd.health, instance_of(UnityHealth))
        assert_that(ssd.needs_replacement, equal_to(False))
        assert_that(ssd.slot_number, equal_to(0))
        assert_that(ssd.name, equal_to("SP A Internal Disk"))
        assert_that(ssd.manufacturer, equal_to(""))
        assert_that(ssd.model, equal_to("Intel DC 3500 Series SSDs M.2"))
        assert_that(ssd.firmware_version, equal_to("G201EM05"))
        assert_that(ssd.emc_part_number, equal_to(
            "INTEL SSDSCKHB120G4M           118000040"))
        assert_that(ssd.emc_serial_number, equal_to("BTWM4CO00FTB"))
        assert_that(ssd.vendor_part_number, equal_to(""))
        assert_that(ssd.vendor_serial_number, equal_to(""))
        assert_that(ssd.parent_storage_processor, instance_of(
            UnityStorageProcessor))


class UnityFanTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        fans = t_unity().get_fan()
        assert_that(len(fans), equal_to(10))

        fan = fans[0]
        assert_that(fan.id, equal_to("dpe_fan_a0"))
        assert_that(fan.health, instance_of(UnityHealth))
        assert_that(fan.parent, instance_of(dict))
        assert_that(fan.slot_number, equal_to(0))
        assert_that(fan.name, equal_to("DPE Cooling Module A0"))
        assert_that(fan.emc_part_number, equal_to("100-542-054-05"))
        assert_that(fan.emc_serial_number, equal_to(""))
        assert_that(fan.manufacturer, equal_to(""))
        assert_that(fan.model, equal_to(""))
        assert_that(fan.vendor_part_number, equal_to(""))
        assert_that(fan.needs_replacement, equal_to(False))
        assert_that(fan.parent_dpe, instance_of(UnityDpe))
        assert_that(fan.parent_dpe.name, equal_to('DPE'))

    @patch_rest
    def test_get_nested_properties(self):
        fans = t_unity().get_fan()
        assert_that(len(fans), equal_to(10))

        fan = fans[0]
        assert_that(fan.parent_dpe.id, equal_to("dpe"))
        assert_that(fan.parent_dpe.name, equal_to("DPE"))
        assert_that(fan.parent_dae.id, equal_to("dae"))
        assert_that(fan.parent_dae.name, equal_to("DAE"))


class UnityLicenseTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        licenses = t_unity().get_license()

        assert_that(len(licenses), equal_to(20))

        license = licenses[0]
        assert_that(license.id, equal_to("ANTIVIRUS"))
        assert_that(license.name, equal_to("ANTIVIRUS"))
        assert_that(license.is_installed, equal_to(True))
        assert_that(license.version, equal_to("1.0"))
        assert_that(license.is_valid, equal_to(True))
        assert_that(license.issued, equal_to("2006-09-08T00:00:00.000Z"))
        assert_that(license.expires, equal_to("2017-06-30T00:00:00.000Z"))
        assert_that(license.is_permanent, equal_to(False))
        assert_that(license.feature, instance_of(UnityFeature))


class UnityFeatureTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        features = t_unity().get_feature()

        assert_that(len(features), equal_to(37))
        feature = features[0]
        assert_that(feature.id, equal_to("ADVANCED_STATIC_ROUTING"))
        assert_that(feature.name, equal_to("ADVANCED_STATIC_ROUTING"))
        assert_that(feature.state, equal_to(FeatureStateEnum.Enabled))
        assert_that(feature.reason, none())
        assert_that(feature.license, none())


class UnityMgmtInterface(TestCase):
    @patch_rest
    def test_get_properties(self):
        interfaces = t_unity().get_mgmt_interface()
        assert_that(len(interfaces), equal_to(1))

        interface = interfaces[0]
        assert_that(interface.id, equal_to('mgmt_ipv4'))
        assert_that(interface.config_mode,
                    equal_to(InterfaceConfigModeEnum.AUTO))
        assert_that(interface.protocol_version,
                    equal_to(IpProtocolVersionEnum.IPv4))
        assert_that(interface.ip_address, equal_to('10.245.101.39'))
        assert_that(interface.netmask, equal_to('255.255.255.0'))
        assert_that(interface.gateway, equal_to('10.245.101.1'))
        assert_that(interface.ethernet_port.id, equal_to('spb_mgmt'))


class UnitySystemCapacity(TestCase):
    @patch_rest
    def test_get_properties(self):
        capacities = t_unity().get_system_capacity()
        assert_that(len(capacities), equal_to(1))

        capacity = capacities[0]
        assert_that(capacity.id, equal_to('0'))
        assert_that(capacity.size_free, equal_to(9496172691456))
        assert_that(capacity.size_total, equal_to(9641664708608))
        assert_that(capacity.size_used, equal_to(145492017152))
        assert_that(capacity.compression_size_saved, equal_to(0))
        assert_that(capacity.compression_percent, equal_to(0))
        assert_that(capacity.compression_ratio, equal_to(1))
        assert_that(capacity.data_reduction_size_saved, equal_to(0))
        assert_that(capacity.data_reduction_percent, equal_to(0))
        assert_that(capacity.data_reduction_ratio, equal_to(1))
        assert_that(capacity.size_subscribed, equal_to(1018980990976))
        assert_that(capacity.tiers[0], instance_of(UnitySystemTierCapacity))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

import six
from hamcrest import assert_that, raises
from hamcrest import equal_to
from storops import exception

from storops.exception import VNXFsNotFoundError, VNXException, raise_if_err
from storops.vnx.enums import VNXProvisionEnum, \
    VNXTieringEnum, VNXSPEnum, VNXRaidType, \
    VNXMigrationRate, VNXPortType, VNXPoolRaidType, VNXCtrlMethod
from storops.vnx.nas_client import NasXmlResponse
from storops_test.vnx.nas_mock import MockXmlPost


class VNXErrorTest(TestCase):
    def test_has_error_with_specific_error(self):
        def f():
            msg = ("SP A: Expansion LUN size must be "
                   "greater than current LUN size. (0x712d8e04)")
            raise_if_err(msg)

        assert_that(f, raises(exception.VNXLunExpandSizeError))

    def test_cg_not_found(self):
        def f():
            output = "Cannot find the consistency group."
            raise_if_err(output)

        assert_that(f, raises(exception.VNXConsistencyGroupNotFoundError))

    def test_snap_not_exists(self):
        def f():
            output = "The specified snapshot does not exist."
            raise_if_err(output)

        assert_that(f, raises(exception.VNXSnapNotExistsError))

    def test_pool_lun_not_exists_multi_line(self):
        def f():
            output = """Could not retrieve the specified (pool lun).
                    The (pool lun) may not exist."""
            raise_if_err(output)

        assert_that(f, raises(exception.VNXLunNotFoundError))

    def test_has_error_regular_string_false(self):
        def f():
            output = ("Cannot unbind LUN because "
                      "it's contained in a Storage Group.")
            raise_if_err(output)

        assert_that(f, raises(exception.VNXLunInStorageGroupError))

    def test_has_error_ev_error(self):
        class ForTest(object):
            pass

        error = ForTest()
        error.where = 'EV_ScsiPipe::_sendCommand() - Sense Data'
        error.why = 'SP A: LUN already exists in the specified storage group.'
        error.who = '@(#)libconnect Revision 7.33.6.2.50 on 1/6/2015 21:54:55'

        def f():
            raise_if_err(error)

        assert_that(f, raises(exception.VNXAluAlreadyAttachedError))

    def test_sp_error_not_supported(self):
        def f():
            out = ('Error returned from the target: 10.244.211.32\n'
                   'CLI commands are not supported by the '
                   'target storage system.')
            raise_if_err(out)

        assert_that(f, raises(exception.VNXNotSupportedError))

    def test_sp_error_time_out(self):
        def f():
            out = ("A network error occurred while "
                   "trying to connect: '10.244.211.33'.\n"
                   "Message : select: The connect timed out.")
            raise_if_err(out)

        assert_that(f, raises(exception.VNXSpNotAvailableError))

    def test_raise_if_err_normal(self):
        raise_if_err('')
        # no raises

    def test_raise_if_err_non_empty(self):
        def f():
            raise_if_err('error msg', msg="error received")

        assert_that(f, raises(VNXException, "error received"))

    def test_raise_if_err_lun_not_found(self):
        def f():
            out = ('Could not retrieve the specified (pool lun). '
                   'The (pool lun) may not exist')
            raise_if_err(out)

        assert_that(f, raises(exception.VNXLunNotFoundError))

    def test_raise_if_err_nas_response_input(self):
        def f():
            resp = NasXmlResponse(MockXmlPost.read_file('fs_not_found.xml'))
            resp.raise_if_err()

        assert_that(f, raises(VNXFsNotFoundError, 'not found'))


class VNXProvisionEnumTest(TestCase):
    def test_get_opt_dedup(self):
        opt = VNXProvisionEnum.get_opt(VNXProvisionEnum.DEDUPED)
        assert_that(' '.join(opt), equal_to('-type Thin -deduplication on'))

    def test_get_opt_thin(self):
        opt = VNXProvisionEnum.get_opt(VNXProvisionEnum.THIN)
        assert_that(' '.join(opt), equal_to('-type Thin'))

    def test_get_opt_thick(self):
        opt = VNXProvisionEnum.get_opt(VNXProvisionEnum.THICK)
        assert_that(' '.join(opt), equal_to('-type NonThin'))

    def test_get_opt_compressed(self):
        opt = VNXProvisionEnum.get_opt(VNXProvisionEnum.COMPRESSED)
        assert_that(' '.join(opt), equal_to('-type Thin'))

    def test_get_opt_not_available(self):
        def f():
            VNXProvisionEnum.get_opt('na')

        assert_that(f, raises(ValueError))


class VNXTieringEnumTest(TestCase):
    def test_get_opt(self):
        opt = VNXTieringEnum.get_opt(VNXTieringEnum.HIGH_AUTO)
        assert_that(
            ' '.join(opt),
            equal_to('-initialTier highestAvailable -tieringPolicy autoTier'))

    def test_get_opt_not_available(self):
        def f():
            VNXTieringEnum.get_opt('na')

        assert_that(f, raises(ValueError))

    def test_invalid_tier_enum(self):
        def f():
            VNXTieringEnum('abc')

        assert_that(f, raises(ValueError, 'not a valid VNXTieringEnum'))

    def test_valid_tier_enum(self):
        auto = VNXTieringEnum('auto')
        assert_that(auto, equal_to(VNXTieringEnum.AUTO))


class VNXSPEnumTest(TestCase):
    def test_from_str(self):
        data = {
            'spa': VNXSPEnum.SP_A,
            'sp': None,
            'sp_a': VNXSPEnum.SP_A,
            'SP b': VNXSPEnum.SP_B,
            'a': VNXSPEnum.SP_A,
            'b': VNXSPEnum.SP_B,
            'cs': VNXSPEnum.CONTROL_STATION,
            'Celerra_CS0_21111': VNXSPEnum.CONTROL_STATION,
            'VPI-24092B': VNXSPEnum.SP_B
        }

        for k, v in six.iteritems(data):
            assert_that(VNXSPEnum.parse(k), equal_to(v),
                        'input: {}'.format(k))

    def test_get_sp_index_err(self):
        def f():
            VNXSPEnum.get_sp_index('abc')

        assert_that(f, raises(ValueError, 'not a valid sp'))

    def test_get_sp_index(self):
        assert_that(VNXSPEnum.get_sp_index('spa'), equal_to('a'))

    def test_sp_value(self):
        assert_that(VNXSPEnum.SP_B.value, equal_to('SP B'))

    def test_index(self):
        assert_that(VNXSPEnum.SP_A.index, equal_to('a'))
        assert_that(VNXSPEnum.SP_B.index, equal_to('b'))


class VNXRaidTypeTest(TestCase):
    def test_from_str(self):
        assert_that(VNXRaidType.from_str('r5'), equal_to(VNXRaidType.RAID5))

    def test_disk_requirement(self):
        assert_that(VNXPoolRaidType.RAID5.min_disk_requirement, equal_to(3))


class VNXMigrationRateTest(TestCase):
    def test_text_type(self):
        assert_that(six.text_type(VNXMigrationRate.HIGH),
                    equal_to('{"VNXMigrationRate": {"value": "high"}}'))


class VNXPortTypeTest(TestCase):
    def test_parse_iqn(self):
        ret = VNXPortType.parse('iqn.1992-04.com.emc:c.a.b')
        assert_that(ret, equal_to(VNXPortType.ISCSI))

    def test_parse_wwn(self):
        ret = VNXPortType.parse('50:06:01:60:B6:E0:16:81:'
                                '50:06:01:68:36:E4:16:81')
        assert_that(ret, equal_to(VNXPortType.FC))


class VNXControlMethodTest(TestCase):
    def test_get_no_ctrl(self):
        no_ctrl = VNXCtrlMethod(method=VNXCtrlMethod.NO_CTRL)
        options = no_ctrl.get_option()
        assert_that(options, equal_to(['-noctrl']))

    def test_get_limit_ctrl(self):
        limit_ctrl = VNXCtrlMethod(method=VNXCtrlMethod.LIMIT_CTRL,
                                   metric='bw',
                                   value=100)
        options = limit_ctrl.get_option()
        assert_that(options, equal_to(['-ctrlmethod', 'limit', '-gmetric',
                                       'bw', '-gval', 100]))

    def test_get_cruise_ctrl(self):
        cruise_ctrl = VNXCtrlMethod(method=VNXCtrlMethod.CRUISE_CTRL,
                                    metric='bw',
                                    value=100,
                                    tolerance=10)
        options = cruise_ctrl.get_option()
        assert_that(options, equal_to(['-ctrlmethod', 'cruise', '-gmetric',
                                       'bw', '-gval', 100, '-gtol', 10]))

    def test_get_fix_ctrl(self):
        fixed_ctrl = VNXCtrlMethod(method=VNXCtrlMethod.FIXED_CTRL,
                                   value=100)
        options = fixed_ctrl.get_option()
        assert_that(options, equal_to(['-ctrlmethod', 'fixed', '-gval', 100]))

    def test_get_invalid_ctrl(self):

        def _inner():
            invalid = VNXCtrlMethod(method='invalid')
            invalid.get_option()

        assert_that(_inner, raises(ValueError,
                                   'Invalid control method specified.'))



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from __future__ import unicode_literals

from storops import HostLUNAccessEnum
from storops.exception import UnityStorageResourceNameInUseError, \
    UnityConsistencyGroupNameInUseError
from storops.unity import enums
from storops.unity.resource import lun as lun_mod
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.replication_session import UnityResourceConfig, \
    UnityReplicationSession
from storops.unity.resource.snap import UnitySnap, UnitySnapList
from storops.unity.resource.storage_resource import UnityStorageResource, \
    UnityStorageResourceList
from storops.unity.resp import RESP_OK

__author__ = 'Peter Wang'

import logging

log = logging.getLogger(__name__)


class UnityConsistencyGroup(UnityStorageResource):
    """
    Create a consistency group (former LUN group).
    """

    @classmethod
    def create(cls, cli, name, description=None, is_repl_dst=None,
               snap_schedule=None, is_snap_schedule_paused=None,
               skip_sync_to_remote_system=None, tiering_policy=None,
               is_compression=None,
               hosts=None, lun_add=None):
        req_body = cls._compose_cg_parameters(cli, name, description,
                                              is_repl_dst,
                                              snap_schedule,
                                              is_snap_schedule_paused,
                                              skip_sync_to_remote_system,
                                              tiering_policy,
                                              is_compression,
                                              hosts,
                                              lun_add)
        resp = cli.type_action(UnityStorageResource().resource_class,
                               'createConsistencyGroup', **req_body)
        try:
            resp.raise_if_err()
        except UnityStorageResourceNameInUseError:
            raise UnityConsistencyGroupNameInUseError()
        except:  # noqa
            raise
        return UnityConsistencyGroup(_id=resp.resource_id, cli=cli)

    @property
    def name(self):
        if hasattr(self, '_name') and self._name is not None:
            name = self._name
        else:
            if not self._is_updated():
                self.update()
            name = self._get_value_by_key('name')
        return name

    @name.setter
    def name(self, new_name):
        self.modify(name=new_name)

    def modify(self, name=None, description=None, is_repl_dst=None,
               snap_schedule=None, is_snap_schedule_paused=None,
               skip_sync_to_remote_system=None, tiering_policy=None,
               is_compression=None, hosts=None,
               lun_add=None, lun_remove=None, host_add=None,
               host_remove=None, lun_modify=None):
        req_body = self._compose_cg_parameters(
            self._cli, name, description,
            is_repl_dst, snap_schedule,
            is_snap_schedule_paused, skip_sync_to_remote_system,
            tiering_policy, is_compression, hosts,
            lun_add, lun_remove, host_add, host_remove, lun_modify)
        resp = self._cli.action(UnityStorageResource().resource_class,
                                self.get_id(), 'modifyConsistencyGroup',
                                **req_body)
        resp.raise_if_err()
        return resp

    @staticmethod
    def _compose_cg_parameters(cli, name=None, description=None,
                               is_repl_dst=None,
                               snap_schedule=None,
                               is_snap_schedule_paused=None,
                               skip_sync_to_remote_system=None,
                               tiering_policy=None,
                               is_compression=None,
                               hosts=None,
                               lun_add=None,
                               lun_remove=None,
                               host_add=None,
                               host_remove=None,
                               lun_modify=None):
        host_access = UnityConsistencyGroup._wrap_host_list(hosts, cli)
        host_access_add = UnityConsistencyGroup._wrap_host_list(host_add, cli)

        lun_add = UnityConsistencyGroup._wrap_lun_list(lun_add)
        lun_remove = UnityConsistencyGroup._wrap_lun_list(lun_remove)

        req_body = cli.make_body(
            name=name,
            description=description,
            replicationParameters=cli.make_body(
                isReplicationDestination=is_repl_dst),
            fastVPParameters=cli.make_body(
                tieringPolicy=tiering_policy),
            dataReductionParameters=cli.make_body(
                isDataReductionEnabled=is_compression
            ),
            snapScheduleParameters=cli.make_body(
                snapSchedule=snap_schedule,
                isSnapSchedulePaused=is_snap_schedule_paused,
                skipSyncToRemoteSystem=skip_sync_to_remote_system,
            ),
            blockHostAccess=cli.make_body(host_access),
            lunAdd=cli.make_body(lun_add),
            lunRemove=cli.make_body(lun_remove),
            addBlockHostAccess=cli.make_body(host_access_add),
            removeBlockHostAccess=cli.make_body(host_remove),
        )
        # Empty host access can be used to wipe the host_access
        if lun_modify is not None:
            lun_modify = cli.make_body(lun_modify, allow_empty=True)
            req_body['lunModify'] = lun_modify

        return req_body

    def modify_lun(self, lun, name=None, size=None, host_access=None,
                   description=None, sp=None, io_limit_policy=None,
                   tiering_policy=None, is_compression=None):

        lun_modify = self._cli.make_body(lun=lun,
                                         name=name,
                                         description=description)
        # `hostAccess` could be empty list which is used to remove all host
        # access
        lun_parameters = lun_mod.prepare_lun_parameters(
            cli=self._cli,
            size=size,
            host_access=host_access,
            sp=sp,
            io_limit_policy=io_limit_policy,
            tiering_policy=tiering_policy,
            is_compression=is_compression)
        if lun_parameters:
            lun_modify['lunParameters'] = lun_parameters

        return self.modify(lun_modify=[lun_modify])

    @staticmethod
    def _wrap_host_list(host_list, cli):
        def make_host_elem(host):
            return cli.make_body(host=host,
                                 accessMask=enums.HostLUNAccessEnum.BOTH)

        if host_list is not None:
            ret = list(map(make_host_elem, host_list))
        else:
            ret = None
        return ret

    @staticmethod
    def _wrap_lun_list(lun_list):
        if lun_list is not None:
            ret = list(map(lambda lun: {'lun': lun}, lun_list))
        else:
            ret = None
        return ret

    def add_lun(self, *lun_list):
        return self.modify(lun_add=lun_list)

    def remove_lun(self, *lun_list):
        return self.modify(lun_remove=lun_list)

    @property
    def existing_lun_ids(self):
        luns = self.luns if self.luns else []
        return {lun.get_id() for lun in luns}

    def _prepare_luns_add(self, lun_list):
        if not lun_list:
            return None
        add_ids = {lun.get_id() for lun in lun_list} - self.existing_lun_ids
        log.debug("Adding LUN(s) '{}' to cg {}".format(add_ids, self.get_id()))
        return [UnityLun(_id=_id, cli=self._cli) for _id in add_ids]

    def _prepare_luns_remove(self, lun_list, is_delta):
        if not lun_list:
            return None
        lun_ids = {lun.get_id() for lun in lun_list}
        if is_delta:
            remove_ids = self.existing_lun_ids & lun_ids
        else:
            remove_ids = self.existing_lun_ids - lun_ids
        log.debug("Removing LUN(s) '{}' from cg '{}'".format(remove_ids,
                                                             self.get_id()))
        return [UnityLun(_id=_id, cli=self._cli) for _id in remove_ids]

    def replace_lun(self, *lun_list):
        """Replaces the exiting LUNs to lun_list."""
        lun_add = self._prepare_luns_add(lun_list)
        lun_remove = self._prepare_luns_remove(lun_list, False)
        return self.modify(lun_add=lun_add, lun_remove=lun_remove)

    def update_lun(self, add_luns=None, remove_luns=None):
        """Updates the LUNs in CG, adding the ones in `add_luns` and removing
        the ones in `remove_luns`"""
        if not add_luns and not remove_luns:
            log.debug("Empty add_luns and remove_luns passed in, "
                      "skip update_lun.")
            return RESP_OK
        lun_add = self._prepare_luns_add(add_luns)
        lun_remove = self._prepare_luns_remove(remove_luns, True)
        return self.modify(lun_add=lun_add, lun_remove=lun_remove)

    def attach_to(self, host=None, access_mask=HostLUNAccessEnum.BOTH):
        """Attaches all member LUNs to the specified host."""
        raise NotImplementedError()

    def detach_from(self, host=None):
        """Detaches all members from host, if None, detach from all hosts."""
        raise NotImplementedError()

    def set_host_access(self, *hosts):
        return self.modify(hosts=hosts)

    def add_host_access(self, *hosts):
        return self.modify(host_add=hosts)

    def remove_host_access(self, *hosts):
        return self.modify(host_remove=hosts)

    def create_snap(self, name=None, description=None, is_auto_delete=None,
                    retention_duration=None):
        clz = UnitySnap
        return clz.create(self._cli, self, name=name, description=description,
                          is_auto_delete=is_auto_delete,
                          retention_duration=retention_duration,
                          is_read_only=None, fs_access_type=None)

    @property
    def snapshots(self):
        clz = UnitySnapList
        snaps = clz(cli=self._cli, storage_resource=self)
        return list(filter(lambda snap: snap.snap_group is None, snaps))

    def check_cg_is_replicated(self):
        if not (self.replication_type is enums.ReplicationTypeEnum.NONE):
            return True

    def replicate_cg_with_dst_resource_provisioning(self,
                                                    max_time_out_of_sync,
                                                    source_luns,
                                                    dst_pool_id,
                                                    dst_cg_name=None,
                                                    remote_system=None):
        """
        Creates a replication session with destination luns provisioning.

        :param max_time_out_of_sync: maximum time to wait before syncing the
            source and destination. Value `-1` means the automatic sync is not
            performed. `0` means it is a sync replication.
        :param source_luns: local Luns, will create same named Luns in
            destination Unity.
        :param dst_pool_id: id of pool to allocate destination Luns.
        :param dst_cg_name: destination cg name.
        :param remote_system: `UnityRemoteSystem` object. The remote system to
            which the replication is being configured. When not specified, it
            defaults to local system.
        :return: created replication session.
        """

        dst_resource = UnityResourceConfig.to_embedded(
            name=dst_cg_name)
        dst_element_configs = []
        for source_lun in source_luns:
            lun_resource_config = UnityResourceConfig.to_embedded(
                pool_id=dst_pool_id,
                is_thin_enabled=source_lun.is_thin_enabled,
                size=source_lun.size_total, request_id=source_lun.id,
                name=source_lun.name,
                is_deduplication_enabled=source_lun.is_data_reduction_enabled,
                is_compression_enabled=source_lun.is_data_reduction_enabled)
            dst_element_configs.append(lun_resource_config)
        result = UnityReplicationSession.create_with_dst_resource_provisioning(
            self._cli, self.get_id(), dst_resource, max_time_out_of_sync,
            remote_system=remote_system,
            dst_resource_element_configs=dst_element_configs)
        return result


class UnityConsistencyGroupList(UnityStorageResourceList):
    type_cg = enums.StorageResourceTypeEnum.CONSISTENCY_GROUP

    def __init__(self, **the_filter):
        the_filter['type'] = self.type_cg
        super(UnityConsistencyGroupList, self).__init__(**the_filter)

    @classmethod
    def get_resource_class(cls):
        return UnityConsistencyGroup

    def _filter(self, item):
        return item.type == self.type_cg



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import inspect
from distutils.version import LooseVersion

import six

from storops.exception import StoropsException, SystemAPINotSupported
from storops.lib.resource import Resource

__author__ = 'Tina Tang'

_op_mapping_ = {}


class Criteria(object):
    def __init__(self, op, sentry, sentry2=None):
        self.op = self.get_op(op)
        self.sentry = sentry
        self.sentry2 = sentry2

    def _support_op(*args):
        """Internal decorator to define an criteria compare operations."""
        def inner(func):
            for one_arg in args:
                _op_mapping_[one_arg] = func
            return func

        return inner

    def test(self, other):
        if other is None:
            return True
        return self.op(self, other)

    @_support_op('=', '==')
    def _eq(self, other):
        return self.sentry == LooseVersion(other)

    @_support_op('<>')
    def _between(self, other):
        return self.sentry < LooseVersion(other) < self.sentry2

    @_support_op('<')
    def _lt(self, other):
        return LooseVersion(other) < self.sentry

    @_support_op('>')
    def _gt(self, other):
        return LooseVersion(other) > self.sentry

    @_support_op('>=')
    def _ge(self, other):
        return LooseVersion(other) >= self.sentry

    @_support_op('<=')
    def _le(self, other):
        return LooseVersion(other) <= self.sentry

    @_support_op('!=')
    def _ne(self, other):
        return LooseVersion(other) != self.sentry

    @classmethod
    def get_op(cls, op_str):
        if op_str not in _op_mapping_:
            raise StoropsException("Invalid version criteria definition.")
        return _op_mapping_[op_str]

    @classmethod
    def parse(cls, criteria_str):
        # Sorted the supported ops by length. The longer one go first.
        support_ops = sorted(_op_mapping_.keys(), key=len, reverse=True)
        for op in support_ops:
            if criteria_str.find(op) != -1:
                fields = criteria_str.split(op)
                sentrys = [LooseVersion(field.strip()) for field in fields if
                           len(field.strip()) != 0]
                return cls(op, *sentrys)
        else:
            raise StoropsException("Invalid version criteria definition.")


class VersionedFuncManager(object):
    __func_mapping__ = {}

    @classmethod
    def add_func(cls, key, versioned_func):
        cls.__func_mapping__.setdefault(key, []).append(versioned_func)

    @classmethod
    def get_support_func(cls, key, version):
        all_funcs = cls.__func_mapping__[key]
        valid_func = [func.func for func in all_funcs if func.support(version)]
        if len(valid_func) == 0:
            raise SystemAPINotSupported(version=version)
        return valid_func[-1]

    @staticmethod
    def get_function_name(func):
        if six.PY2:
            if hasattr(func, "im_class"):
                return "%s.%s" % (func.im_class, func.__name__)
            else:
                return "%s.%s" % (func.__module__, func.__name__)
        else:
            return "%s.%s" % (func.__module__, func.__qualname__)


class VersionedFunc(object):
    def __init__(self, criteria, func):
        self.criteria = criteria
        self.func = func

    def support(self, version):
        return self.criteria.test(version)


class Version(object):
    INIT_METHOD = 1
    CLASS_METHOD = 2
    INSTANCE_METHOD = 3
    STATIC_METHOD = 4
    METHOD = 5

    def __init__(self, criteria_string, version_getter):
        self.criteria = Criteria.parse(criteria_string)
        self.version_getter = version_getter

    def __call__(self, obj):
        if inspect.isclass(obj):
            return self._cls_decorator(obj)
        else:
            return self._func_decorator(obj)

    def _func_decorator(self, func):
        func_key = VersionedFuncManager.get_function_name(func)
        versioned_func = VersionedFunc(self.criteria, func)
        VersionedFuncManager.add_func(func_key, versioned_func)

        @six.wraps(func)
        def _inner(*args, **kwargs):
            version = self.version_getter(Version.METHOD,
                                          *args, **kwargs)

            inner_func = VersionedFuncManager.get_support_func(
                func_key, version)
            return inner_func(*args, **kwargs)

        return _inner

    def _cls_decorator(self, cls):
        mro_dict = [o.__dict__ for o in inspect.getmro(cls)]
        for attr_name, attr_val in inspect.getmembers(cls):
            if (not inspect.ismethod(attr_val) and
                    not inspect.isfunction(attr_val)):
                # Only handle functions
                continue

            if self._is_classmethod(attr_name, mro_dict):
                setattr(cls, attr_name,
                        self._handle_classmethod(attr_val))

            elif self._is_staticmethod(attr_name, mro_dict):
                setattr(cls, attr_name, self._handle_staticmethod(attr_val))

            elif attr_name == '__init__':
                setattr(cls, attr_name, self._handle_init_method(attr_val))

        return cls

    @staticmethod
    def _is_classmethod(attr_name, mro_dict):
        for mro in mro_dict:
            if attr_name in mro and isinstance(mro[attr_name],
                                               classmethod):
                return True
        return False

    @staticmethod
    def _is_staticmethod(attr_name, mro_dict):
        for mro in mro_dict:
            if attr_name in mro and isinstance(mro[attr_name],
                                               staticmethod):
                return True
        return False

    def _handle_staticmethod(self, func):
        _inner = self._handle_method(Version.STATIC_METHOD, func)
        return staticmethod(_inner)

    def _handle_classmethod(self, func,):
        def _inner(cls, *args, **kwargs):
            self._test_support(Version.CLASS_METHOD,
                               cls, *args, **kwargs)
            return func(*args, **kwargs)
        return classmethod(_inner)

    def _handle_init_method(self, func):
        return self._handle_method(Version.INIT_METHOD, func)

    def _handle_method(self, dec_type, func):
        def _inner(*args, **kwargs):
            self._test_support(dec_type, *args, **kwargs)
            return func(*args, **kwargs)

        return _inner

    def _test_support(self, dec_type, *args, **kwargs):
        version = self.version_getter(dec_type, *args, **kwargs)
        if not self.criteria.test(version):
            raise SystemAPINotSupported(version=version)


def get_version(dec_type, *args, **kwargs):
    if len(args) > 0:
        obj = args[0]
        if (isinstance(obj, Resource) and
                (dec_type == Version.INSTANCE_METHOD or
                 dec_type == Version.METHOD)):
            return obj.system_version
    return _get_version_from_client_in_args(*args, **kwargs)


def is_valid_client_instance(inst):
    return inst.__class__.__name__ in ('UnityClient', 'CliClient')


def _get_version_from_client_in_args(*args, **kwargs):
    for parm in list(args) + list(kwargs.values()):
        if is_valid_client_instance(parm):
            return parm.system_version
    return None


def version(criteria_string):
    """Version requirement decorator

    This decorator can be used to decorate either a subclass of Resource or
    a memthod of Resource subclass. It supports version validation and
    versioned methods

    1. Decorate a class
    The decorated class can not be instantiated and the static methods and
    class methods of it will not be able to be invoked when the version
    criteria is not satisfied. The version information is retrieved from the
    cli parameter which could be a UnityClient or CliClient.

    For example:

    >>>  @version('>3.0')
    >>>  class OneResource(Resource):
    >>>      def __init__(self, cli)
    >>>          super(OneResource, self).__init__()
    >>>          self._cli = cli

    2. Decorate a method
    The decorated method will not be able to be invoked if the version
    criteria is not satisfied. For instance method, the version information
    is retrieved from the instance itself. For staticmethod, constructor or
    classmethod, the version information is retrieved from the cli parameter
    which could be a UnityClient or CliClient.
    The versioned method is supported with using this decorator. Multiple
    methods can be defined with a same name after decorating it with different
    versions criterias. When the method is invoked, the method that satisfying
    the version will be used. Note that the '# noqa' is needed to get rid of
    pep8 error.

    For example:

    >>>  class TwoResource(Resource):
    >>>      def __init__(self, cli)
    >>>          super(OneResource, self).__init__()
    >>>          self._cli = cli
    >>>
    >>>      @staticmethod
    >>>      @version('<2.0')
    >>>      def static_method(a)
    >>>          pass
    >>>
    >>>      @version('>2.0')
    >>>      def method(a, b)
    >>>          return '>2.0'
    >>>
    >>>      @version('<=2.0')  # noqa
    >>>      def method(a, b)
    >>>          return '<=2.0'

    :param criteria_string: The version criteria expression. The supported
         comparator includes: <, <=, =, ==, !=, >, >=, <>(between)
    """
    return Version(criteria_string, get_version)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re
from datetime import datetime, timedelta
from functools import partial
from operator import is_not

import bitmath
import dateutil.parser
import six

from storops.vnx import enums

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)

NAs = ['N/A', 'Unbound', 'None', 'Disabled',
       'This disk does not belong to a RAIDGroup']


def to_bool(str_input):
    ret = False
    if str_input.strip().lower() in ('yes', 'true', 'enabled', 'on'):
        ret = True
    return ret


def to_hex(number):
    if number is not None:
        h = hex(number)
        if h.endswith('L'):
            h = h[:-1]
    else:
        h = None
    return h


def to_wwn(str_input):
    str_input = str_input.upper()
    if ':' not in str_input:
        items = [str_input[i:i + 2] for i in range(0, len(str_input), 2)]
        ret = ':'.join(items)
    else:
        ret = str_input
    return ret


def to_int_arr(inputs):
    if inputs is not None:
        if isinstance(inputs, six.string_types):
            inputs = re.split(',| ', inputs)
        ints = map(to_int, inputs)
        ret = list(filter(partial(is_not, None), ints))
    else:
        ret = []
    return ret


def to_int_str_map(str_input):
    ret = {}
    for pair in re.findall(r'(\w+):\s*(\w+)', str_input):
        ret[to_int(pair[0])] = pair[1].strip()
    return ret


def to_int_int_map(str_input):
    ret = {}
    for pair in re.findall(r'(\w+):\s*(\w+)', str_input):
        ret[to_int(pair[0])] = to_int(pair[1])
    return ret


def arr_to_str(int_arr, sep=None):
    if sep is None:
        sep = ','
    return sep.join((six.text_type(i) for i in int_arr))


def to_str_arr(int_arr):
    return [six.text_type(i) for i in int_arr]


def to_float(value):
    ret = None
    if value is not None:
        if value.endswith('%'):
            value = value[:-1]
        try:
            ret = float(value)
        except ValueError:
            pass
    return ret


def to_int(value):
    ret = None
    if value is not None:
        try:
            if isinstance(value, six.integer_types):
                ret = value
            elif len(value) > 0:
                ret = int(value)
        except ValueError:
            if value not in NAs:
                log.warn('cannot convert "{}" to int.'.format(value))
    return ret


def to_alu_hlu_map(input_str):
    """Converter for alu hlu map

    Convert following input into a alu -> hlu map:
    Sample input:

    ```
      HLU Number     ALU Number
      ----------     ----------
        0               12
        1               23
    ```

    ALU stands for array LUN number
    hlu stands for host LUN number
    :param input_str: raw input from naviseccli
    :return: alu -> hlu map
    """
    ret = {}
    if input_str is not None:
        pattern = re.compile(r'(\d+)\s*(\d+)')
        for line in input_str.split('\n'):
            line = line.strip()
            if len(line) == 0:
                continue
            matched = re.search(pattern, line)
            if matched is None or len(matched.groups()) < 2:
                continue
            else:
                hlu = matched.group(1)
                alu = matched.group(2)
                ret[int(alu)] = int(hlu)
    return ret


def to_disk_indices(value):
    """Convert following input to disk indices

    Sample input:

    ```
    Disks:
    Bus 0 Enclosure 0 Disk 9
    Bus 1 Enclosure 0 Disk 12
    Bus 1 Enclosure 0 Disk 9
    Bus 0 Enclosure 0 Disk 4
    Bus 0 Enclosure 0 Disk 7
    ```

    :param value: disk list
    :return: disk indices in list
    """
    ret = []
    p = re.compile(r'Bus\s+(\w+)\s+Enclosure\s+(\w+)\s+Disk\s+(\w+)')
    if value is not None:
        for line in value.split('\n'):
            line = line.strip()
            if len(line) == 0:
                continue
            matched = re.search(p, line)
            if matched is None or len(matched.groups()) < 3:
                continue
            else:
                ret.append('{}_{}_{}'.format(*matched.groups()))
    return ret


def vnx_time_to_date(value):
    return datetime.utcfromtimestamp(value - 2177452800)


def to_datetime(value):
    return dateutil.parser.parse(value)


def to_time_delta(value):
    hours, minutes, seconds = map(float, value.split(':'))
    return timedelta(hours=hours, minutes=minutes, seconds=seconds)


def _to_enum(enum_class):
    def _enum_converter(value):
        return enum_class.parse(value)

    return _enum_converter


to_sp_enum = _to_enum(enums.VNXSPEnum)
to_mirror_view_recovery_policy = _to_enum(enums.VNXMirrorViewRecoveryPolicy)
to_mirror_view_sync_rate = _to_enum(enums.VNXMirrorViewSyncRate)
to_mirror_group_recovery_policy = _to_enum(enums.VNXMirrorGroupRecoveryPolicy)
to_raid_type = _to_enum(enums.VNXRaidType)
to_pool_raid_type = _to_enum(enums.VNXPoolRaidType)
to_port_type = _to_enum(enums.VNXPortType)
to_migration_rate_enum = _to_enum(enums.VNXMigrationRate)


def to_raid_type_list(value):
    return [to_raid_type(x) for x in value.split()]


def boolean_to_str(value, true_str='true', false_str='false'):
    if value:
        ret = true_str
    else:
        ret = false_str
    return ret


def url_to_host(url):
    """convert a url to a host (ip or domain)

    :param url: url string
    :returns: host: domain name or ipv4/v6 address
    :rtype: str
    :raises: ValueError: given an illegal url that without a ip or domain name
    """

    regex_url = r"([a-z][a-z0-9+\-.]*://)?" + \
                r"([a-z0-9\-._~%!$&'()*+,;=]+@)?" + \
                r"([a-z0-9\-._~%]+" + \
                r"|\[[a-z0-9\-._~%!$&'()*+,;=:]+\])?" + \
                r"(:(?P<port>[0-9]+))?"

    m = re.match(regex_url, url, re.IGNORECASE)
    if m and m.group(3):
        return url[m.start(3): m.end(3)]
    else:
        raise ValueError("URL without a valid host or ip")


def parse_host_address(addr):
    """
    parse host address to get domain name or ipv4/v6 address,
    cidr prefix and net mask code string if given a subnet address

    :param addr:
    :type addr: str
    :return: parsed domain name/ipv4 address/ipv6 address,
             cidr prefix if there is,
             net mask code string if there is
    :rtype: (string, int, string)
    """

    if addr.startswith('[') and addr.endswith(']'):
        addr = addr[1:-1]

    parts = addr.split('/')
    if len(parts) == 1:
        return parts[0], None, None
    if len(parts) > 2:
        raise ValueError("Illegal host address")
    else:
        domain_or_ip, prefix = parts
        prefix = int(prefix)
        if re.match(r"^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$", domain_or_ip):
            return domain_or_ip, prefix, ipv4_prefix_to_mask(prefix)
        elif ':' in domain_or_ip:
            return domain_or_ip, prefix, ipv6_prefix_to_mask(prefix)
        else:
            return domain_or_ip, None, None


def ipv4_prefix_to_mask(prefix):
    """
    ipv4 cidr prefix to net mask

    :param prefix: cidr prefix , rang in (0, 32)
    :type prefix: int
    :return: dot separated ipv4 net mask code, eg: 255.255.255.0
    :rtype: str
    """
    if prefix > 32 or prefix < 0:
        raise ValueError("invalid cidr prefix for ipv4")
    else:
        mask = ((1 << 32) - 1) ^ ((1 << (32 - prefix)) - 1)
        eight_ones = 255  # 0b11111111
        mask_str = ''
        for i in range(0, 4):
            mask_str = str(mask & eight_ones) + mask_str
            mask = mask >> 8
            if i != 3:
                mask_str = '.' + mask_str
        return mask_str


def ipv6_prefix_to_mask(prefix):
    """
    ipv6 cidr prefix to net mask

    :param prefix: cidr prefix, rang in (0, 128)
    :type prefix: int
    :return: comma separated ipv6 net mask code,
             eg: ffff:ffff:ffff:ffff:0000:0000:0000:0000
    :rtype: str
    """
    if prefix > 128 or prefix < 0:
        raise ValueError("invalid cidr prefix for ipv6")
    else:
        mask = ((1 << 128) - 1) ^ ((1 << (128 - prefix)) - 1)
        f = 15  # 0xf or 0b1111
        hex_mask_str = ''
        for i in range(0, 32):
            hex_mask_str = format((mask & f), 'x') + hex_mask_str
            mask = mask >> 4
            if i != 31 and i & 3 == 3:
                hex_mask_str = ':' + hex_mask_str
        return hex_mask_str


def mb_to_gb(mib):
    return bitmath.MiB(mib).to_GiB().value


def block_to_gb(block):
    return bitmath.Byte(block * 512).to_GiB().value


def to_minute(minute_str):
    ret = None
    if minute_str:
        m = re.match(r"00:(\d+):00.000$", minute_str)
        if m:
            ret = int(m.group(1))
    return ret


def from_minute(minute_in_int):
    if minute_in_int:
        if 1 <= minute_in_int <= 60:
            ret = "00:%02d:00.000" % minute_in_int
        else:
            raise ValueError("Value for minute must be 1 - 60.")
    else:
        ret = None
    return ret


def to_hour(hour_str):
    ret = None
    if hour_str:
        m = re.match(r"(\d+):00:00.000$", hour_str)
        if m:
            ret = int(m.group(1))
    return ret


def from_hour(hour_in_int):
    if hour_in_int:
        if 1 <= hour_in_int <= 24:
            ret = "%02d:00:00.000" % hour_in_int
        else:
            raise ValueError("Value for hour must be 1 - 24.")
    else:
        ret = None
    return ret



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import equal_to, assert_that, instance_of, raises

from storops.exception import UnityNfsAlreadyEnabledError, \
    UnityResourceNotFoundError
from storops.unity.resource.interface import UnityFileInterfaceList
from storops.unity.resource.nas_server import UnityNasServer
from storops.unity.resource.nfs_server import UnityNfsServer, \
    UnityNfsServerList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


class UnityNfsServerTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        server = UnityNfsServer('nfs_2', t_rest())
        assert_that(server.id, equal_to('nfs_2'))
        assert_that(server.nfs_v4_enabled, equal_to(True))
        assert_that(server.is_secure_enabled, equal_to(False))
        assert_that(server.is_extended_credentials_enabled, equal_to(False))
        assert_that(server.nas_server, instance_of(UnityNasServer))
        assert_that(server.file_interfaces,
                    instance_of(UnityFileInterfaceList))
        assert_that(str(server.credentials_cache_ttl), equal_to('0:15:00'))

    @patch_rest
    def test_get_all(self):
        servers = UnityNfsServerList(cli=t_rest())
        assert_that(len(servers), equal_to(1))

    @patch_rest
    def test_create_success(self):
        server = UnityNfsServer.create(t_rest(), 'nas_5', nfs_v4_enabled=True)
        assert_that(server.id, equal_to('nfs_3'))
        assert_that(server.nfs_v4_enabled, equal_to(True))

    @patch_rest
    def test_create_existed(self):
        def f():
            UnityNfsServer.create(t_rest(), 'nas_5', nfs_v4_enabled=False)

        assert_that(f, raises(UnityNfsAlreadyEnabledError, 'already enabled'))

    @patch_rest
    def test_delete_success(self):
        server = UnityNfsServer(_id='nfs_3', cli=t_rest())
        resp = server.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_not_found(self):
        def f():
            server = UnityNfsServer(_id='nfs_5', cli=t_rest())
            server.delete()

        assert_that(f, raises(UnityResourceNotFoundError))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from hamcrest import assert_that, equal_to, has_items, has_item, none

from storops.vnx.enums import VNXSPEnum, VNXProvisionEnum, VNXTieringEnum, \
    VNXRaidType

__author__ = 'Cedric Zhuang'


def verify_lun_0(lun):
    assert_that(lun.lun_id, equal_to(0))
    assert_that(lun.name, equal_to('File_CS0_21132_0_d7'))
    assert_that(lun.state, equal_to('Ready'))
    assert_that(lun.current_owner, equal_to(VNXSPEnum.SP_A))
    assert_that(lun.default_owner, equal_to(VNXSPEnum.SP_A))
    assert_that(lun.wwn, equal_to(
        '60:06:01:60:12:60:3D:00:95:63:38:87:9D:69:E5:11'))
    assert_that(lun.operation, equal_to('None'))
    assert_that(lun.pool_name, equal_to('Pool4File'))
    assert_that(lun.is_thin_lun, equal_to(False))
    assert_that(lun.is_compressed, equal_to(False))
    assert_that(lun.is_dedup, equal_to(False))
    assert_that(lun.is_private, equal_to(False))
    assert_that(lun.tier, equal_to(VNXTieringEnum.HIGH_AUTO))
    assert_that(lun.provision, equal_to(VNXProvisionEnum.THICK))
    assert_that(lun.user_capacity_gbs, equal_to(500.0))
    assert_that(lun.consumed_capacity_gbs, equal_to(512.249))
    assert_that(lun.existed, equal_to(True))
    assert_that(lun.primary_lun, none())
    assert_that(lun.is_snap_mount_point, equal_to(False))


def verify_raid0(rg):
    assert_that(rg.raid_group_id, equal_to(0))
    assert_that(rg.raid_group_type, equal_to(VNXRaidType.RAID5))
    assert_that(rg.state, equal_to('Valid_luns'))
    assert_that(rg.disks.index,
                has_items('0_0_A0', '0_0_A1', '0_0_A2', '0_0_A3', '0_0_A4'))
    assert_that(rg.list_of_luns, has_item(63868))
    assert_that(len(rg.list_of_luns), equal_to(16))
    assert_that(rg.max_number_of_disks, equal_to(16))
    assert_that(rg.max_number_of_luns, equal_to(256))
    assert_that(rg.raw_capacity_blocks, equal_to(4502487040))
    assert_that(rg.logical_capacity_blocks, equal_to(4502478848))
    assert_that(rg.free_capacity_blocks_non_contiguous,
                equal_to(3744083968))
    assert_that(rg.free_contiguous_group_of_unbound_segments,
                equal_to(1749913216))
    assert_that(rg.defrag_expand_priority, equal_to('N/A'))
    assert_that(rg.percent_defragmented, none())
    assert_that(rg.percent_expanded, none())
    assert_that(rg.disk_expanding_onto, equal_to('N/A'))
    assert_that(rg.lun_expansion_enabled, equal_to(False))
    assert_that(rg.legal_raid_types[0], equal_to(VNXRaidType.RAID5))


def verify_raid_1(rg):
    assert_that(rg.raid_group_id, equal_to(0))
    assert_that(rg.raid_group_type, equal_to(VNXRaidType.UNBOUND))
    assert_that(len(rg.legal_raid_types), equal_to(4))
    assert_that(rg.legal_raid_types[0], equal_to(VNXRaidType.RAID5))
    assert_that(rg.legal_raid_types[1], equal_to(VNXRaidType.RAID6))
    assert_that(rg.legal_raid_types[2], equal_to(VNXRaidType.RAID0))
    assert_that(rg.legal_raid_types[3], equal_to(VNXRaidType.RAID10))


def verify_pool_0(pool):
    assert_that(pool.name, equal_to('Pool4File'))
    assert_that(pool.pool_id, equal_to(0))
    assert_that(pool.state, equal_to('Ready'))
    assert_that(pool.status, equal_to('OK(0x0)'))
    assert_that(pool.fast_cache, equal_to(False))
    assert_that(pool.available_capacity_gbs, equal_to(17314.501))
    assert_that(pool.consumed_capacity_gbs, equal_to(540.303))
    assert_that(pool.total_subscribed_capacity_gbs, equal_to(540.053))
    assert_that(pool.user_capacity_gbs, equal_to(17854.805))
    assert_that(pool.current_operation, equal_to('None'))
    assert_that(pool.current_operation_percent_completed, equal_to(0.0))
    assert_that(pool.current_operation_state, equal_to('N/A'))
    assert_that(pool.current_operation_status, equal_to('N/A'))
    assert_that(pool.luns, equal_to([0]))
    assert_that(pool.percent_full_threshold, equal_to(70.0))
    assert_that(pool.existed, equal_to(True))


def verify_disk_4_0_e8(disk):
    assert_that(disk.index_string, equal_to('bus 4 enclosure 0 disk E8'))
    assert_that(disk.bus, equal_to('4'))
    assert_that(disk.enclosure, equal_to('0'))
    assert_that(disk.disk, equal_to('E8'))
    assert_that(disk.vendor_id, equal_to("SEAGATE"))
    assert_that(disk.product_id, equal_to("ST990080 CLAR900"))
    assert_that(disk.product_revision, equal_to("CS18"))
    assert_that(disk.lun,
                has_items(62, 63, 306, 324, 63871, 1200, 326, 355))
    assert_that(disk.type[62], equal_to('RAID5'))
    assert_that(disk.state, equal_to('Enabled'))
    assert_that(disk.hot_spare, equal_to('NO'))
    assert_that(disk.prct_rebuilt[62], equal_to(99))
    assert_that(disk.prct_bound[62], equal_to(92))
    assert_that(disk.serial_number, equal_to('6XS2EAKG'))
    assert_that(disk.sectors, equal_to('90701824 (44288)'))
    assert_that(disk.capacity, equal_to(840313))
    assert_that(disk.private[63], equal_to(335872))
    assert_that(disk.bind_signature, equal_to('0x22c2ed, 0, 56'))
    assert_that(disk.hard_read_errors, equal_to(1))
    assert_that(disk.hard_write_errors, equal_to(2))
    assert_that(disk.soft_read_errors, equal_to(3))
    assert_that(disk.soft_write_errors, equal_to(4))
    assert_that(disk.read_retries, none())
    assert_that(disk.write_retries, none())
    assert_that(disk.remapped_sectors, equal_to('N/A'))
    assert_that(disk.number_of_reads, equal_to(33646149))
    assert_that(disk.number_of_writes, equal_to(5655147))
    assert_that(disk.number_of_luns, equal_to(8))
    assert_that(disk.raid_group_id, equal_to(5))
    assert_that(disk.clariion_part_number, equal_to('DG118032758'))
    assert_that(disk.request_service_time, equal_to('N/A'))
    assert_that(disk.read_requests, equal_to(33646149))
    assert_that(disk.write_requests, equal_to(5655147))
    assert_that(disk.kbytes_read, equal_to(1858798320))
    assert_that(disk.kbytes_written, equal_to(876138634))
    assert_that(disk.stripe_boundary_crossing, none())
    assert_that(disk.drive_type, equal_to('SAS'))
    assert_that(disk.clariion_tla_part_number, equal_to('005049207PWR'))
    assert_that(disk.user_capacity, equal_to(0))
    assert_that(disk.idle_ticks, equal_to(118477646))
    assert_that(disk.busy_ticks, equal_to(1288999))
    assert_that(disk.current_speed, equal_to('6Gbps'))
    assert_that(disk.maximum_speed, equal_to('6Gbps'))
    assert_that(disk.queue_max, none())
    assert_that(disk.queue_avg, none())
    assert_that(disk.prct_idle, equal_to(7))
    assert_that(disk.prct_busy, equal_to(8))
    assert_that(disk.hardware_power_savings_qualified, equal_to(True))
    assert_that(disk.hardware_power_savings_eligible, equal_to(True))
    assert_that(disk.power_savings_state, equal_to('Full Power'))
    assert_that(disk.current_power_savings_log_timestamp, equal_to('N/A'))
    assert_that(disk.spinning_ticks, none())
    assert_that(disk.standby_ticks, none())
    assert_that(disk.number_of_spin_ups, none())
    assert_that(disk.arrivals_with_nonzero_queue, equal_to(8354391))
    assert_that(disk.high_sum_of_seeks, equal_to(1211052417356144))
    assert_that(disk.idle_ticks_spa, equal_to(59428411))
    assert_that(disk.idle_ticks_spb, equal_to(59049235))
    assert_that(disk.busy_ticks_spa, equal_to(456326))
    assert_that(disk.busy_ticks_spb, equal_to(832673))
    assert_that(disk.queue_length, equal_to(72099433))



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import storops.unity.resource.nas_server
import storops.unity.resource.pool
from storops.unity.enums import NodeEnum
from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Jay Xu, Cedric Zhuang'

LOG = logging.getLogger(__name__)


class UnityStorageProcessor(UnityResource):
    def create_nas_server(self, name, pool=None, is_repl_dst=None,
                          multi_proto=None, tenant=None):
        if pool is None:
            pool_list_clz = storops.unity.resource.pool.UnityPoolList
            pool = self._get_unity_rsc(pool_list_clz).first_item
        nas_server_clz = storops.unity.resource.nas_server.UnityNasServer
        return nas_server_clz.create(self._cli, name, self, pool,
                                     is_repl_dst=is_repl_dst,
                                     multi_proto=multi_proto,
                                     tenant=tenant)

    def to_node_enum(self):
        if self.name in ['SP A']:
            value = 0
        elif self.name in ['SP B']:
            value = 1
        else:
            # NodeEnum.Unknown
            value = 2989
        return NodeEnum.parse(value)


class UnityStorageProcessorList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityStorageProcessor



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Cedric Zhuang'


class UnityStorageResource(UnityResource):
    @classmethod
    def get(cls, cli, _id=None):
        if not isinstance(_id, cls):
            ret = cls(_id=_id, cli=cli)
        else:
            ret = _id
        return ret

    def action(self, action_name, **kwargs):
        return self._cli.action(self.resource_class,
                                self.get_id(),
                                action_name,
                                **kwargs)

    def modify_fs(self, **kwargs):
        return self.action('modifyFilesystem', **kwargs)


class UnityStorageResourceList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityStorageResource



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

import mock

from hamcrest import assert_that, equal_to, instance_of, raises, none, calling

from storops.exception import UnityShareOnCkptSnapError, \
    UnityDeleteAttachedSnapError, UnityResourceNotFoundError, \
    UnitySnapAlreadyPromotedException, UnityException, \
    UnityThinCloneNotAllowedError, UnityCGMemberActionNotSupportError
from storops.unity.enums import FilesystemSnapAccessTypeEnum, \
    SnapCreatorTypeEnum, SnapStateEnum, NFSTypeEnum, CIFSTypeEnum
from storops.unity.resource.filesystem import UnityFileSystem
from storops.unity.resource.host import UnityHost
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.snap import UnitySnap, UnitySnapList
from storops.unity.resource.storage_resource import UnityStorageResource
from storops.unity.resource.port import UnityIoLimitPolicy

from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


class UnitySnapTest(TestCase):
    @patch_rest
    def test_properties(self):
        snap = UnitySnap(_id=171798691852, cli=t_rest())
        assert_that(snap.existed, equal_to(True))
        assert_that(snap.state, equal_to(SnapStateEnum.READY))
        assert_that(snap.name, equal_to('esa_nfs1_2016-03-15_10:56:29'))
        assert_that(snap.is_system_snap, equal_to(False))
        assert_that(snap.is_modifiable, equal_to(False))
        assert_that(snap.is_read_only, equal_to(False))
        assert_that(snap.is_modified, equal_to(False))
        assert_that(snap.is_auto_delete, equal_to(True))
        assert_that(snap.size, equal_to(5368709120))
        assert_that(str(snap.creation_time),
                    equal_to('2016-03-15 02:57:27.092000+00:00'))
        assert_that(snap.storage_resource, instance_of(UnityStorageResource))
        assert_that(snap.creator_type,
                    equal_to(SnapCreatorTypeEnum.USER_CUSTOM))
        assert_that(snap.access_type,
                    equal_to(FilesystemSnapAccessTypeEnum.CHECKPOINT))
        assert_that(snap.is_cg_snap(), equal_to(False))

    @patch_rest
    def test_get_all(self):
        snaps = UnitySnapList(cli=t_rest())
        assert_that(snaps[0].existed, equal_to(True))
        assert_that(len(snaps), equal_to(3))

    @patch_rest
    def test_create_snap_success(self):
        snap = UnitySnap(_id='171798691884', cli=t_rest())
        sos = snap.create_snap(name='snap_over_snap')
        assert_that(sos.name, equal_to('snap_over_snap'))
        assert_that(sos.existed, equal_to(True))

    @patch_rest
    def test_delete_snap(self):
        snap = UnitySnap(_id='171798691885', cli=t_rest())
        resp = snap.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_create_nfs_share_type_error(self):
        def f():
            snap = UnitySnap(cli=t_rest(), _id='171798691852')
            snap.create_nfs_share('sns1')

        assert_that(f, raises(UnityShareOnCkptSnapError, 'is a checkpoint'))

    @patch_rest
    def test_create_nfs_share_success(self):
        snap = UnitySnap(cli=t_rest(), _id='171798691896')
        share = snap.create_nfs_share('sns1')
        assert_that(share.snap, equal_to(snap))
        assert_that(share.name, equal_to('sns1'))
        assert_that(share.type, equal_to(NFSTypeEnum.NFS_SNAPSHOT))

    @patch_rest
    def test_create_cifs_share_success(self):
        snap = UnitySnap(cli=t_rest(), _id='171798691899')
        share = snap.create_cifs_share('sns2')
        assert_that(share.snap, equal_to(snap))
        assert_that(share.name, equal_to('sns2'))
        assert_that(share.type, equal_to(CIFSTypeEnum.CIFS_SNAPSHOT))

    @patch_rest
    def test_filesystem_snap(self):
        snap = UnitySnap(cli=t_rest(), _id='171798691852')
        fs = snap.filesystem
        assert_that(fs, instance_of(UnityFileSystem))
        assert_that(fs.storage_resource, equal_to(snap.storage_resource))
        assert_that(snap.lun, none())

    @patch_rest
    def test_lun_snap(self):
        snap = UnitySnap(cli=t_rest(), _id='38654705785')
        lun = snap.lun
        assert_that(lun, instance_of(UnityLun))
        assert_that(snap.filesystem, none())

    @patch_rest
    def test_copy_snap_success(self):
        snap = UnitySnap(cli=t_rest(), _id='38654705785')
        snap = snap.copy('s3')
        assert_that(snap.existed, equal_to(True))
        assert_that(snap, instance_of(UnitySnap))

    @patch_rest
    def test_destroying_snap_existed(self):
        snap = UnitySnap(cli=t_rest(), _id='171798691953')
        assert_that(snap.existed, equal_to(False))

    @patch_rest
    def test_not_found_snap_existed(self):
        snap = UnitySnap(cli=t_rest(), _id='12345')
        assert_that(snap.existed, equal_to(False))

    @patch_rest
    def test_delete_not_exist_snap(self):
        snap = UnitySnap(_id='38654705844', cli=t_rest())
        assert_that(lambda: snap.delete(), raises(UnityResourceNotFoundError))

    @patch_rest
    def test_delete_attached_snap(self):
        snap = UnitySnap(_id='38654705845', cli=t_rest())
        assert_that(lambda: snap.delete(),
                    raises(UnityDeleteAttachedSnapError))

    @patch_rest
    def test_delete_attached_snap_with_force(self):
        snap = UnitySnap(_id='38654705845', cli=t_rest())

        def fake_detach_from(host):
            snap._id = '171798691885'

        with mock.patch.object(UnitySnap, 'detach_from',
                               side_effect=fake_detach_from):
            resp = snap.delete(even_attached=True)
            assert_that(snap._id, equal_to('171798691885'))
            assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_attach_snap_success(self):
        snap = UnitySnap(_id='38654705676', cli=t_rest())
        host = UnityHost(_id="Host_12", cli=t_rest())
        resp = snap.attach_to(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_attach_second_snap_success(self):
        snap = UnitySnap(_id='38654705670', cli=t_rest())
        host = UnityHost(_id="Host_19", cli=t_rest())
        resp = snap.attach_to(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_detach_snap_success(self):
        snap = UnitySnap(_id='38654705676', cli=t_rest())
        host = UnityHost(_id="Host_12", cli=t_rest())
        resp = snap.detach_from(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_attach_snap_already_attached(self):
        def f():
            snap = UnitySnap(_id='38654705676', cli=t_rest())
            host = UnityHost(_id="Host_13", cli=t_rest())
            snap.attach_to(host)

        assert_that(f, raises(UnitySnapAlreadyPromotedException, "promoted"))

    @patch_rest
    def test_is_cg_snap(self):
        snap = UnitySnap(_id='85899345930', cli=t_rest())
        assert_that(snap.is_cg_snap(), equal_to(True))

    @patch_rest
    def test_get_member_snap_not_found(self):
        def f():
            snap = UnitySnap(_id='85899345930', cli=t_rest())
            lun = UnityLun(cli=t_rest(), _id='sv_3342')
            snap.get_member_snap(lun)

        assert_that(f, raises(ValueError, 'no instance'))

    @patch_rest
    def test_get_member_snap_found(self):
        snap_group = UnitySnap(_id='85899345930', cli=t_rest())
        lun = UnityLun(cli=t_rest(), _id='sv_3338')
        snap = snap_group.get_member_snap(lun)
        assert_that(snap.snap_group.get_id(), equal_to(snap_group.get_id()))
        assert_that(snap.storage_resource.get_id(), equal_to('res_19'))
        assert_that(snap.lun.get_id(), equal_to(lun.get_id()))

    @patch_rest
    def test_thin_clone(self):
        snap = UnitySnap(_id='38654705847', cli=t_rest(version='4.2.0'))
        policy = UnityIoLimitPolicy(_id='qp_2', cli=t_rest())
        clone = snap.thin_clone(name='test_thin_clone',
                                description='This is description.',
                                io_limit_policy=policy)
        assert_that(clone.id, equal_to('sv_4678'))

    @patch_rest
    def test_thin_clone_auto_delete(self):
        snap = UnitySnap(_id='38654705848', cli=t_rest(version='4.2.0'))

        def _inner():
            snap.thin_clone(name='test_thin_clone',
                            description='This is description.',
                            io_limit_policy=None)
        assert_that(_inner, raises(UnityException, "Error Code:0x670166b"))

    @patch_rest
    def test_snap_restore(self):
        snap = UnitySnap(cli=t_rest(), _id='38654705785')
        backup = snap.restore(backup="backup_snap", delete_backup=True)
        assert_that(backup.id, equal_to("38654700002"))

    @patch_rest
    def test_snap_restore_auto_delete_false(self):
        snap = UnitySnap(cli=t_rest(), _id='38654705785')
        backup = snap.restore(backup="backup_snap")
        assert_that(backup.id, equal_to("38654700002"))

    @patch_rest
    def test_thinclone_of_member_snap_not_support(self):
        snap = UnitySnap(cli=t_rest(version='4.3'), _id='38654707282')
        assert_that(calling(snap.thin_clone).with_args('not-support'),
                    raises(UnityCGMemberActionNotSupportError))

    @patch_rest
    def test_thinclone_of_thick_lun_snap_not_allowed(self):
        snap = UnitySnap(cli=t_rest(version='4.3'), _id='38654707273')
        assert_that(calling(snap.thin_clone).with_args('not-allowed'),
                    raises(UnityThinCloneNotAllowedError))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import threading
from logging.handlers import RotatingFileHandler
import sys
import os
from os.path import join, dirname, abspath
from fasteners import process_lock

import errno
from time import sleep

from retryz import retry

from storops.exception import StoropsException
from storops.lib.common import get_lock_file
from storops_test.utils import PersistedDict

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


def log_folder():
    folder = join(dirname(abspath(__file__)), '..', 'logs')
    if not os.path.exists(folder):
        try:
            os.makedirs(folder)
        except OSError as ex:
            if ex.errno != errno.EEXIST:
                raise
    return folder


def setup_log():
    fmt_str = ('%(asctime)s [%(levelname)s] %(process)d '
               '%(name)s - %(message)s')
    level = logging.DEBUG
    filename = join(log_folder(), 'comp_test.log')

    root = logging.getLogger()
    root.setLevel(level)

    ch = logging.StreamHandler(sys.stdout)
    ch.setFormatter(logging.Formatter(fmt_str))
    ch.setLevel(logging.INFO)

    fh = RotatingFileHandler(filename, maxBytes=10 * 1024 ** 2, backupCount=20)
    fh.setFormatter(logging.Formatter(fmt_str))

    root.addHandler(ch)
    root.addHandler(fh)


class ResourceManager(object):
    def __init__(self, name):
        self._names = self._init_names()
        self._name = name
        self._add_worker()
        log.info('initialize fixture "{}"'.format(self._name))

    @property
    def name(self):
        return self._name

    def setup(self):
        """ Detail setup should be done in this method

        :return: nothing
        """
        pass

    def clean_up(self):
        log.info('wait for all workers to exit.')
        while self._remove_worker():
            # wait for all test cases to complete
            sleep(1)

        log.info('start fixture {} clean up.'.format(self.clz_name))

    def _init_names(self):
        data_file = '{}_names'.format(self.clz_name)
        return PersistedDict(data_file, list)

    @property
    def clz_name(self):
        return self.__class__.__name__

    def _add_worker(self):
        workers = self._names['worker']
        _id = threading.current_thread().ident
        if _id not in workers:
            workers.append(_id)
        self._names['worker'] = workers

    def _remove_worker(self):
        workers = self._names['worker']
        _id = threading.current_thread().ident
        if _id in workers:
            workers.remove(_id)
        self._names['worker'] = workers
        return len(workers)

    def has_name(self, rsc_type, key=None):
        names = self._names[rsc_type]
        if names:
            if key is None:
                ret = True
            else:
                ret = key in names
        else:
            ret = False
        return ret

    def _pop_name(self, rsc_type):
        names = self._names[rsc_type]
        ret = names.pop()
        self._names[rsc_type] = names
        return ret

    def add_name(self, rsc_type, name=None):
        names = self._names[rsc_type]
        n = len(names)

        if name is None:
            name = '{}_{}_{}'.format(self._name, rsc_type, n)

        if name not in names:
            names.append(name)
            self._names[rsc_type] = names
        return name

    @staticmethod
    @retry(on_error=ValueError, wait=5, limit=20)
    def until(obj, checker):
        obj.update()
        if not checker(obj):
            raise ValueError

    @classmethod
    def until_existed(cls, obj):
        return cls.until(obj, lambda x: x.existed)

    @classmethod
    def until_not_existed(cls, obj):
        return cls.until(obj, lambda x: not x.existed)

    ######################
    # has_xxx_name methods
    def has_snap_name(self, name=None):
        return self.has_name('snap', name)

    def has_lun_name(self, name=None):
        return self.has_name('lun', name)

    def has_mirror_name(self, name=None):
        return self.has_name('mirror', name)

    def has_pool_name(self, name=None):
        return self.has_name('pool', name)

    def has_fs_name(self, name=None):
        return self.has_name('fs', name)

    def has_cg_name(self, name=None):
        return self.has_name('cg', name)

    def has_nas_server_name(self, name=None):
        return self.has_name('nas_server', name)

    def has_cifs_share_name(self, name=None):
        return self.has_name('cifs_share', name)

    def has_cifs_server_name(self, name=None):
        return self.has_name('cifs_server', name)

    def has_nfs_share_name(self, name=None):
        return self.has_name('nfs_share', name)

    def has_sg_name(self, name=None):
        return self.has_name('sg', name)

    ######################
    # add_xxx_name methods
    def add_snap_name(self, name=None):
        return self.add_name('snap', name)

    def add_lun_name(self, name=None):
        return self.add_name('lun', name)

    def add_mirror_name(self, name=None):
        return self.add_name('mirror', name)

    def add_pool_name(self, name=None):
        return self.add_name('pool', name)

    def add_fs_name(self, name=None):
        return self.add_name('fs', name)

    def add_nas_server_name(self, name=None):
        return self.add_name('nas_server', name)

    def add_cifs_share_name(self, name=None):
        return self.add_name('cifs_share', name)

    def add_cifs_server_name(self, name=None):
        return self.add_name('cifs_server', name)

    def add_nfs_share_name(self, name=None):
        return self.add_name('nfs_share', name)

    def add_sg_name(self, name=None):
        return self.add_name('sg', name)

    def add_cg_name(self, name=None):
        return self.add_name('cg', name)


def is_jenkins():
    return 'jenkins' in os.path.abspath(__file__)


def setup_fixture(request, fixture_clz):
    fixture = fixture_clz()

    @inter_process_locked('{}.lck'.format(fixture.name))
    def _setup():
        try:
            fixture.setup()
        except StoropsException:
            log.exception('fixture setup failed.')

    def _clean_up():
        log.info('tear down general fixture.')
        if fixture:
            fixture.clean_up()

    request.addfinalizer(_clean_up)
    _setup()
    return fixture


def inter_process_locked(name):
    return process_lock.interprocess_locked(get_lock_file(name))



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, only_contains, instance_of, raises
from hamcrest import equal_to
from storops.unity.resource.tenant import UnityTenant, UnityTenantList
from storops.unity.resource.host import UnityHostList
from storops.unity.resource.nas_server import UnityNasServerList
from storops.exception import UnityTenantNameInUseError, SystemAPINotSupported
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Tina Tang'


class UnityTenantTest(TestCase):
    @patch_rest()
    def test_create_tenant(self):
        UnityTenant.create(t_rest('4.1.0'), 'test',
                           uuid='173ca6c3-5952-427d-82a6-df88f49e3926',
                           vlans=[3])

    @patch_rest()
    def test_create_tenant_unsupported(self):
        def f():
            UnityTenant.create(t_rest('4.0.0'), 'test',
                               uuid='173ca6c3-5952-427d-82a6-df88f49e3926',
                               vlans=[3])
        assert_that(f, raises(SystemAPINotSupported))

    @patch_rest()
    def test_create_tenant_failed_name_inused(self):
        def do():
            UnityTenant.create(t_rest('4.1.0'), 'dup',
                               vlans=[3])

        assert_that(do, raises(UnityTenantNameInUseError,
                               'The specified tenant name is already in use'))

    @patch_rest()
    def test_tenant_modify_vlans(self):
        tenant = UnityTenant('tenant_1', cli=t_rest('4.1.0'))
        tenant.modify(vlans=[4])

    @patch_rest()
    def test_tenant_modify_name(self):
        tenant = UnityTenant('tenant_1', cli=t_rest('4.1.0'))
        tenant.modify(name='new_name')

    @patch_rest()
    def test_properties(self):
        tenant = UnityTenant('tenant_1', cli=t_rest('4.1.0'))
        assert_that(tenant.name, equal_to('T1'))
        assert_that(tenant.uuid,
                    equal_to('173ca6c3-5952-427d-82a6-df88f49e3926'))
        assert_that(tenant.vlans, instance_of(list))
        assert_that(tenant.vlans, only_contains(1, 3))
        assert_that(len(tenant.hosts), equal_to(1))
        assert_that(tenant.hosts, instance_of(UnityHostList))
        assert_that(tenant.nas_servers, instance_of(UnityNasServerList))
        assert_that(len(tenant.nas_servers), equal_to(1))

    @patch_rest()
    def test_get_all(self):
        tenants = UnityTenantList(cli=t_rest('4.1.0'))
        assert_that(len(tenants), equal_to(3))

    @patch_rest()
    def test_query_tenant_with_vlans(self):
        tenants = UnityTenantList(cli=t_rest('4.1.0'),
                                  vlans=[1, 3])
        assert_that(len(tenants), equal_to(1))

    @patch_rest()
    def test_delete_tenant(self):
        tenant = UnityTenant('tenant_1', cli=t_rest())
        tenant.delete()

    @patch_rest()
    def test_delete_tenant_with_hosts(self):
        tenant = UnityTenant('tenant_2', cli=t_rest())
        tenant.delete(delete_hosts=True)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, raises

from storops_test.vnx.cli_mock import patch_cli, t_cli
from storops.exception import VNXSnapError, VNXSnapNotExistsError, \
    VNXDeleteAttachedSnapError, VNXRestoreSnapToMismatchLunError, \
    VNXSnapNameInUseError
from storops.vnx.resource.snap import VNXSnap, VNXSnapList

__author__ = 'Cedric Zhuang'


class VNXSnapTest(TestCase):
    @patch_cli
    def test_properties(self):
        snap = VNXSnap('gan_snap', t_cli())
        assert_that(snap.name, equal_to('gan_snap'))
        assert_that(snap.description, equal_to('gan snap'))
        assert_that(snap.creation_time, equal_to('05/24/13 20:06:12'))
        assert_that(snap.last_modify_time, equal_to('07/23/14 12:28:42'))
        assert_that(snap.last_modified_by, equal_to('N/A'))
        assert_that(snap.source_luns, equal_to([57]))
        assert_that(snap.source_cg, equal_to('N/A'))
        assert_that(snap.primary_luns, equal_to([57]))
        assert_that(snap.state, equal_to('Ready'))
        assert_that(snap.status, equal_to('OK(0x0)'))
        assert_that(snap.allow_read_write, equal_to(True))
        assert_that(snap.modified, equal_to(True))
        assert_that(snap.attached_luns, equal_to([]))
        assert_that(snap.allow_auto_delete, equal_to(True))
        assert_that(snap.expiration_date, equal_to('Never'))
        assert_that(snap.existed, equal_to(True))

    @patch_cli
    def test_get_all(self):
        snaps = VNXSnap.get(t_cli())
        assert_that(len(snaps), equal_to(47))

    @patch_cli
    def test_get_by_name(self):
        snap = VNXSnap.get(t_cli(), name='gan_snap')
        assert_that(snap.creation_time, equal_to('05/24/13 20:06:12'))

    @patch_cli(output='snap_-list_-detail_error.txt')
    def test_get_not_found(self):
        snap = VNXSnap.get(t_cli(), name='xxx')
        assert_that(snap.existed, equal_to(False))

    @patch_cli
    def test_copy_snap(self):
        def f():
            src = VNXSnap.get(t_cli(), name='123')
            src.copy('456')

        assert_that(f, raises(VNXSnapError, 'Cannot copy'))

    @patch_cli
    def test_modify_snap(self):
        snap = VNXSnap(cli=t_cli(), name='s1')
        snap.modify(new_name='s2', allow_rw=True)
        assert_that(snap._name, equal_to('s2'))

    @patch_cli
    def test_modify_snap_failed(self):
        snap = VNXSnap(cli=t_cli(), name='s2')
        try:
            snap.modify(new_name='s1')
            self.fail('should have raise an exception.')
        except VNXSnapError:
            assert_that(snap._name, equal_to('s2'))

    @patch_cli
    def test_delete_snap(self):
        def f():
            snap = VNXSnap(cli=t_cli(), name='s3')
            snap.delete()

        assert_that(f, raises(VNXSnapNotExistsError,
                              'Cannot destroy the snapshot'))

    @patch_cli
    def test_delete_snap_attached(self):
        def f():
            snap = VNXSnap(cli=t_cli(), name='s4')
            snap.delete()

        assert_that(f, raises(VNXDeleteAttachedSnapError, 'is attached'))

    @patch_cli
    def test_get_by_res(self):
        snaps = VNXSnapList(cli=t_cli(), res=3)
        assert_that(len(snaps), equal_to(2))

    @patch_cli
    def test_restore_snap(self):
        snap = VNXSnap(cli=t_cli(), name='s5')
        snap.restore(res_id='540')
        # No exception if succeed.

    @patch_cli
    def test_restore_snap_to_mismatch_lun(self):
        def f():
            snap = VNXSnap(cli=t_cli(), name='s5')
            snap.restore(res_id='541')

        assert_that(f, raises(VNXRestoreSnapToMismatchLunError,
                              'associated with different primary LUNs'))

    @patch_cli
    def test_restore_snap_new_backup_snap_exist(self):
        def f():
            snap = VNXSnap(cli=t_cli(), name='s5')
            snap.restore(res_id='540', backup_snap='a-existing-snap')

        assert_that(f, raises(VNXSnapNameInUseError,
                              'snapshot name is already in use'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from functools import wraps

import six

import storops.unity.resource.system
import storops.unity.resource.type_resource
from storops.connection.connector import UnityRESTConnector
from storops.exception import UnityResourceNotSupportedError
from storops.lib.common import instance_cache, EnumList
from storops.lib.metric import PerfManager
from storops.unity.enums import UnityEnum, UnityEnumList
from storops.unity.resource import UnityResource, UnityResourceList
from storops.unity.resp import RestResponse

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


def wrap_not_supported(func):
    @wraps(func)
    def _wrap(*args, **kwargs):
        try:
            ret = func(*args, **kwargs)
        except UnityResourceNotSupportedError:
            ret = RestResponse(inputs="")
        return ret

    return _wrap


class UnityClient(PerfManager):
    def __init__(self, ip, username, password, port=443, verify=False,
                 retries=None, cache_interval=0):
        super(UnityClient, self).__init__()
        self.ip = ip
        self._rest = UnityRESTConnector(ip, port=port, user=username,
                                        password=password,
                                        verify=verify,
                                        retries=retries,
                                        cache_interval=cache_interval)
        self._system_version = None

    @wrap_not_supported
    def get_all(self, type_name, base_fields=None, the_filter=None,
                nested_fields=None):
        """Get the resource by resource id.

        :param nested_fields: nested resource fields
        :param base_fields: fields of this resource
        :param the_filter: dictionary of filter like `{'name': 'abc'}`
        :param type_name: Resource type. For example, pool, lun, nasServer.
        :return: List of resource class objects
        """
        fields = self.get_fields(type_name, base_fields, nested_fields)
        the_filter = self.dict_to_filter_string(the_filter)

        url = '/api/types/{}/instances'.format(type_name)

        resp = self.rest_get(url, fields=fields, filter=the_filter)
        ret = resp
        while resp.has_next_page:
            resp = self.rest_get(url, fields=fields, filter=the_filter,
                                 page=resp.next_page)
            ret.entries.extend(resp.entries)
        return ret

    @classmethod
    def dict_to_filter_string(cls, the_filter):
        def _get_non_list_value(k, v):
            if isinstance(v, six.string_types):
                r = '{} eq "{}"'.format(k, v)
            elif isinstance(v, UnityEnum):
                r = '{} eq {}'.format(k, v.value[0])
            elif isinstance(v, UnityResource):
                r = '{} eq "{}"'.format(k, v.get_id())
            else:
                r = '{} eq {}'.format(k, v)
            return r

        if the_filter:
            items = []
            for key in sorted(the_filter.keys()):
                value = the_filter[key]
                if value is None:
                    continue
                if isinstance(value, (list, tuple, UnityEnumList)):
                    list_ret = ' or '.join([_get_non_list_value(key, item)
                                            for item in value])
                    items.append(list_ret)
                else:
                    items.append(_get_non_list_value(key, value))
            if items:
                ret = ' and '.join(items)
            else:
                ret = None
        else:
            ret = None
        return ret

    def rest_get(self, url, fields=None, **params):
        if fields is None:
            fields = []
        params['fields'] = ','.join(map(str, sorted(fields)))
        url = self.assemble_url(url, **params)
        return RestResponse(self._rest.get(url))

    def rest_post(self, url, body=None, files=None, **params):
        url = self.assemble_url(url, **params)
        return RestResponse(self._rest.post(url, files=files, body=body))

    def rest_delete(self, url, body=None, **params):
        url = self.assemble_url(url, **params)
        return RestResponse(self._rest.delete(url, body=body))

    @classmethod
    def assemble_url(cls, url, **params):
        if not url.startswith('/'):
            url = '/{}'.format(url)
        if 'compact' not in params:
            params['compact'] = True
        param_list = []
        for key in sorted(params.keys()):
            if params[key] is None:
                continue
            param_list.append('{}={}'.format(key, params[key]))
        url = '{}?{}'.format(url, '&'.join(param_list))
        return url

    @instance_cache
    def _get_type_resource(self, type_name):
        type_clz = storops.unity.resource.type_resource.UnityType
        return type_clz(type_name, self)

    def get_fields(self, type_name, base_fields=None, nested_fields=None):
        if base_fields is not None:
            ret = base_fields
        else:
            unity_type = self._get_type_resource(type_name)
            ret = unity_type.fields
        if nested_fields is not None:
            if isinstance(nested_fields, six.text_type):
                nested_fields = tuple([nested_fields])
            ret = ret + nested_fields
        return ret

    def get_doc(self, clz):
        return UnityDoc.get_doc(self, clz)

    @wrap_not_supported
    def get(self, type_name, obj_id, base_fields=None, nested_fields=None):
        """Get the resource by resource id.

        :param nested_fields: nested resource fields.
        :param type_name: Resource type. For example, pool, lun, nasServer.
        :param obj_id: Resource id
        :param base_fields: Resource fields to return
        :return: List of tuple [(name, res_inst)]
        """
        base_fields = self.get_fields(type_name, base_fields, nested_fields)
        url = '/api/instances/{}/{}'.format(type_name, obj_id)
        return self.rest_get(url, fields=base_fields)

    def post(self, type_name, **kwargs):
        url = '/api/types/{}/instances'.format(type_name)
        body = self.make_body(kwargs)
        return self.rest_post(url, body)

    def action(self, type_name, obj_id, action, **kwargs):
        base_url = '/api/instances/{}/{}/action/{}'
        url = base_url.format(type_name, obj_id, action)
        url_params = {}
        if 'async_mode' in kwargs:
            async_mode = kwargs['async_mode']
            del kwargs['async_mode']
            if async_mode:
                url_params['timeout'] = 0
        body = self.make_body(kwargs, allow_empty=True)
        return self.rest_post(url, body, **url_params)

    def modify(self, type_name, obj_id, **kwargs):
        return self.action(type_name, obj_id, 'modify', **kwargs)

    def type_action(self, type_name, action, **kwargs):
        url = '/api/types/{}/action/{}'.format(type_name, action)
        body = self.make_body(kwargs)
        return self.rest_post(url, body)

    def delete(self, type_name, _id, **kwargs):
        url = '/api/instances/{}/{}'.format(type_name, _id)
        url_params = {'compact': True}
        if 'async_mode' in kwargs:
            async_mode = kwargs['async_mode']
            del kwargs['async_mode']
            if async_mode:
                url_params['timeout'] = 0
        body = self.make_body(kwargs)
        return self.rest_delete(url, body, **url_params)

    @classmethod
    def _is_empty(cls, value):
        if isinstance(value, (dict, tuple, list)) and len(value) == 0:
            ret = True
        else:
            ret = False
        return ret

    @classmethod
    def make_body(cls, value=None, allow_empty=False, **kwargs):
        if value is None and kwargs:
            value = kwargs
        if isinstance(value, dict):
            ret = {}
            for k, v in value.items():
                v = cls.make_body(v, allow_empty=allow_empty)
                if v is not None and (allow_empty or not cls._is_empty(v)):
                    ret[k] = v
        elif isinstance(value, (list, tuple, UnityResourceList,
                                UnityEnumList)):
            ret = [cls.make_body(v, allow_empty=allow_empty) for v in value]
        elif isinstance(value, UnityEnum):
            ret = value.index
        elif isinstance(value, UnityResource):
            ret = {'id': value.get_id()}
        else:
            ret = value
        return ret

    def set_system_version(self, version):
        self._system_version = version

    @property
    def system_version(self):
        if self._system_version is None:
            clz = storops.unity.resource.system.UnityBasicSystemInfo
            self._system_version = clz.get(cli=self).software_version
        return self._system_version


class UnityDoc(object):
    def __init__(self, cli, clz):
        self._cli = cli

        if issubclass(clz, UnityResourceList):
            clz = clz.get_resource_class()
        elif issubclass(clz, EnumList):
            clz = clz.get_enum_class()
        self._clz = clz

    @classmethod
    def get_doc(cls, cli, clz):
        return UnityDoc(cli, clz).doc

    @property
    def doc(self):
        if issubclass(self._clz, UnityResource):
            ret = self._get_unity_resource_doc()
        elif issubclass(self._clz, UnityEnum):
            ret = self._get_unity_enum_doc()
        else:
            raise ValueError(
                'get_doc not support {}.'.format(self._clz.__name__))
        return ret

    def _get_unity_enum_doc(self):
        docs = self._header
        docs.append('Members:')
        docs.append('--------')
        props = []
        for index in sorted(self._clz.indices()):
            doc = self._get_doc(value=index)
            props.append((index, doc))
        docs += self.format_prop(props,
                                 header=('Enum Index:', 'Description:'))
        return '\n'.join(docs)

    def _get_unity_resource_doc(self):
        docs = self._header
        docs.append('Properties:')
        docs.append('-----------')
        props = []
        for name in self._clz().property_names():
            field = self._clz.get_property_label(name)
            if field:
                doc = self._get_doc(field=field)
                props.append((name, doc))
        docs += self.format_prop(props,
                                 header=('Property Name:', 'Description:'))
        return '\n'.join(docs)

    @property
    def _header(self):
        title = self._clz.__name__
        ret = [title, '=' * len(title)]
        desc = self._get_doc()
        if desc:
            ret.append(desc)
        ret.append('')
        return ret

    @property
    @instance_cache
    def rsc_name(self):
        if issubclass(self._clz, UnityResource):
            ret = self._clz().resource_class
        elif issubclass(self._clz, UnityEnum):
            ret = self._clz.__name__
        else:
            ret = self._clz
        return ret

    @classmethod
    def format_prop(cls, props, header=None):
        if header is not None:
            props.insert(0, header)

        ret = []
        if props:
            max_char_counts = cls.get_column_max_len(props)
            fmt_str = cls.get_fmt_str(max_char_counts)
            for prop in props:
                ret.append(fmt_str.format(*map(str, prop)).strip())
        return ret

    @staticmethod
    def get_column_max_len(array):
        max_char_count = []
        if array:
            column_size = len(array[0])
            for i in range(column_size):
                max_len = 0
                for prop in array:
                    max_len = max(max_len, len(str(prop[i])))
                max_char_count.append(max_len)
        return max_char_count

    @staticmethod
    def get_fmt_str(max_char_counts, padding=2):
        out = []
        for count in max_char_counts:
            out.append('{{:{}}}'.format(count + padding))
        return ''.join(out)

    def _get_doc(self, field=None, value=None):
        # noinspection PyProtectedMember
        unity_type = self._cli._get_type_resource(self.rsc_name)
        ret = None
        if field is not None:
            if unity_type.attributes:
                for attr in unity_type.attributes:
                    if field == attr.get('name'):
                        ret = attr.get('description')
                        break
        elif value is not None:
            if unity_type.attributes:
                for attr in unity_type.attributes:
                    if value == attr.get('initialValue'):
                        ret = attr.get('description')
                        break
        else:
            ret = unity_type.description

        if ret:
            ret = ret.strip()
        return ret



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, has_items, starts_with

from storops.unity.resource.type_resource import UnityType
from storops_test.unity.rest_mock import patch_rest, t_rest

__author__ = 'Cedric Zhuang'


class UnityTypeTest(TestCase):
    @patch_rest
    def test_get_metric_type(self):
        t = UnityType(_id='metric', cli=t_rest())
        assert_that(t.name, equal_to('metric'))
        assert_that(t.existed, equal_to(True))

    @patch_rest
    def test_get_lun_type(self):
        t = UnityType(_id='lun', cli=t_rest())
        assert_that(t.fields, has_items('creationTime', 'name', 'snapCount'))

    def test_get_type_fields(self):
        t = UnityType(_id='type')
        assert_that(t.fields_str, starts_with(
            'name,description,documentation,type,attributes.name'))

    @patch_rest
    def test_fields_str(self):
        t = UnityType(_id='lun', cli=t_rest())
        assert_that(t.fields_str, starts_with('auSize,creationTime,'))
